# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
Jobclass to execute python scripts and jupyter notebooks
"""

import os
from typing import Optional

from pyiron_base.jobs.job.generic import GenericJob
from pyiron_base.storage.datacontainer import DataContainer

__author__ = "Jan Janssen"
__copyright__ = (
    "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "production"
__date__ = "Sep 1, 2017"


class ScriptJob(GenericJob):
    """
    The ScriptJob class allows to submit Python scripts and Jupyter notebooks to the pyiron job management system.

    Args:
        project (ProjectHDFio): ProjectHDFio instance which points to the HDF5 file the job is stored in
        job_name (str): name of the job, which has to be unique within the project

    Simple example:
        Step 1. Create the notebook to be submitted, for ex. 'example.ipynb', and save it -- Can contain any code like:
                ```
                import json
                with open('script_output.json','w') as f:
                    json.dump({'x': [0,1]}, f)  # dump some data into a JSON file
                ```

        Step 2. Create the submitter notebook, for ex. 'submit_example_job.ipynb', which submits 'example.ipynb' to the
                pyiron job management system, which can have the following code:
                ```
                from pyiron_base import Project
                pr = Project('scriptjob_example')  # save the ScriptJob in the 'scriptjob_example' project
                scriptjob = pr.create.job.ScriptJob('scriptjob')  # create a ScriptJob named 'scriptjob'
                scriptjob.script_path = 'example.ipynb'  # specify the PATH to the notebook you want to submit.
                ```

        Step 3. Check the job table to get details about 'scriptjob' by using:
                ```
                pr.job_table()
                ```

        Step 4. If the status of 'scriptjob' is 'finished', load the data from the JSON file into the
                'submit_example_job.ipynb' notebook by using:
                ```
                import json
                with open(scriptjob.working_directory + '/script_output.json') as f:
                    data = json.load(f)  # load the data from the JSON file
                ```

    More sophisticated example:
        The script in ScriptJob can also be more complex, e.g. running its own pyiron calculations.
        Here we show how it is leveraged to run a multi-core atomistic calculation.

        Step 1. 'example.ipynb' can contain pyiron_atomistics code like:
                ```
                from pyiron_atomistics import Project
                pr = Project('example')
                job = pr.create.job.Lammps('job')  # we name the job 'job'
                job.structure = pr.create.structure.ase_bulk('Fe')  # specify structure

                # Optional: get an input value from 'submit_example_job.ipynb', the notebook which submits
                #   'example.ipynb'
                number_of_cores = pr.data.number_of_cores
                job.server.cores = number_of_cores

                job.run()  # run the job

                # save a custom output, that can be used by the notebook 'submit_example_job.ipynb'
                job['user/my_custom_output'] = 16
                ```

        Step 2. 'submit_example_job.ipynb', can then have the following code:
                ```
                from pyiron_base import Project
                pr = Project('scriptjob_example')  # save the ScriptJob in the 'scriptjob_example' project
                scriptjob = pr.create.job.ScriptJob('scriptjob')  # create a ScriptJob named 'scriptjob'
                scriptjob.script_path = 'example.ipynb'  # specify the PATH to the notebook you want to submit.
                                                         # In this example case, 'example.ipynb' is in the same
                                                         # directory as 'submit_example_job.ipynb'

                # Optional: to submit the notebook to a queueing system
                number_of_cores = 1  # number of cores to be used
                scriptjob.server.cores = number_of_cores
                scriptjob.server.queue = 'cmfe'  # specify the queue to which the ScriptJob job is to be submitted
                scriptjob.server.run_time = 120  # specify the runtime limit for the ScriptJob job in seconds

                # Optional: save an input, such that it can be accessed by 'example.ipynb'
                pr.data.number_of_cores = number_of_cores
                pr.data.write()

                # run the ScriptJob job
                scriptjob.run()
                ```

        Step 3. Check the job table by using:
                ```
                pr.job_table()
                ```
                in addition to containing details on 'scriptjob', the job_table also contains the details of the child
                'job/s' (if any) that were submitted within the 'example.ipynb' notebook.

        Step 4. Reload and analyse the child 'job/s': If the status of a child 'job' is 'finished', it can be loaded
                into the 'submit_example_job.ipynb' notebook using:
                ```
                job = pr.load('job')  # remember in Step 1., we wanted to run a job named 'job', which has now
                                      # 'finished'
                ```
                this loads 'job' into the 'submit_example_job.ipynb' notebook, which can be then used for analysis,
                ```
                job.output.energy_pot[-1]  # via the auto-complete
                job['user/my_custom_output']  # the custom output, directly from the hdf5 file
                ```

    Attributes:

        attribute: job_name

            name of the job, which has to be unique within the project

        .. attribute:: status

            execution status of the job, can be one of the following [initialized, appended, created, submitted, running,
                                                                      aborted, collect, suspended, refresh, busy, finished]

        .. attribute:: job_id

            unique id to identify the job in the pyiron database

        .. attribute:: parent_id

            job id of the predecessor job - the job which was executed before the current one in the current job series

        .. attribute:: master_id

            job id of the master job - a meta job which groups a series of jobs, which are executed either in parallel or in
            serial.

        .. attribute:: child_ids

            list of child job ids - only meta jobs have child jobs - jobs which list the meta job as their master

        .. attribute:: project

            Project instance the jobs is located in

        .. attribute:: project_hdf5

            ProjectHDFio instance which points to the HDF5 file the job is stored in

        .. attribute:: job_info_str

            short string to describe the job by it is job_name and job ID - mainly used for logging

        .. attribute:: working_directory

            working directory of the job is executed in - outside the HDF5 file

        .. attribute:: path

            path to the job as a combination of absolute file system path and path within the HDF5 file.

        .. attribute:: version

            Version of the hamiltonian, which is also the version of the executable unless a custom executable is used.

        .. attribute:: executable

            Executable used to run the job - usually the path to an external executable.

        .. attribute:: library_activated

            For job types which offer a Python library pyiron can use the python library instead of an external executable.

        .. attribute:: server

            Server object to handle the execution environment for the job.

        .. attribute:: queue_id

            the ID returned from the queuing system - it is most likely not the same as the job ID.

        .. attribute:: logger

            logger object to monitor the external execution and internal pyiron warnings.

        .. attribute:: restart_file_list

            list of files which are used to restart the calculation from these files.

        .. attribute:: job_type

            Job type object with all the available job types: ['ExampleJob', 'ParallelMaster', 'ScriptJob',
                                                               'ListMaster']

        .. attribute:: script_path

            the absolute path to the python script
    """

    def __init__(self, project, job_name):
        super(ScriptJob, self).__init__(project, job_name)
        self.__version__ = "0.1"
        self.__hdf_version__ = "0.2.0"
        self.__name__ = "Script"
        self._script_path = None
        self.input = DataContainer(table_name="custom_dict")
        self._enable_mpi4py = False
        # Set job_with_calculate_function flag to true to use run_static() to execute the python function generated by
        # the job with its arguments job.get_calculate_function(**job.calculate_kwargs) without calling the old
        # interface with write_input() and collect_output(). Finally, the output dictionary is stored in the HDF5 file
        # using self.save_output(output_dict, shell_output)
        self._job_with_calculate_function = True

    @property
    def script_path(self) -> str:
        """
        Python script path

        Returns:
            str: absolute path to the python script
        """
        return self._script_path

    @script_path.setter
    def script_path(self, path: str) -> None:
        """
        Python script path

        Args:
            path (str): relative or absolute path to the python script or a corresponding notebook
        """
        if isinstance(path, str):
            self._script_path = os.path.normpath(
                os.path.join(os.path.abspath(os.path.curdir), path)
            )
            self.executable = self._executable_command(
                working_directory=self.working_directory,
                script_path=self._script_path,
                enable_mpi4py=self._enable_mpi4py,
                cores=self.server.cores,
            )
            if self._enable_mpi4py:
                self.executable._mpi = True
        else:
            raise TypeError(
                "path should be a string, but ", path, " is a ", type(path), " instead."
            )

    def collect_logfiles(self) -> None:
        """
        Compatibility function - but no log files are being collected
        """
        pass

    def disable_mpi4py(self) -> None:
        """
        Disables the usage of mpi4py for parallel execution.
        """
        self._enable_mpi4py = False

    def enable_mpi4py(self) -> None:
        """
        Enable the usage of mpi4py for parallel execution.
        """
        self._enable_mpi4py = True

    def _from_dict(self, obj_dict: dict, version: Optional[str] = None) -> None:
        """
        Load job attributes from a dictionary.

        Args:
            obj_dict (dict): The dictionary containing the job attributes.
            version (str): The version of the job object.
        """
        super()._from_dict(obj_dict=obj_dict)
        if "parallel" in obj_dict["input"].keys():
            self._enable_mpi4py = obj_dict["input"]["parallel"]
        path = obj_dict["input"]["path"]
        if path is not None:
            self.script_path = path
        if "custom_dict" in obj_dict["input"].keys():
            self.input.update(obj_dict["input"]["custom_dict"])

    def get_input_parameter_dict(self) -> dict:
        """
        Get an hierarchical dictionary of input files. On the first level the dictionary is divided in file_to_create
        and files_to_copy. Both are dictionaries use the file names as keys. In file_to_create the values are strings
        which represent the content which is going to be written to the corresponding file. In files_to_copy the values
        are the paths to the source files to be copied.

        Returns:
            dict: hierarchical dictionary of input files
        """
        input_file_dict = super().get_input_parameter_dict()
        if self._script_path is not None:
            files_to_copy_dict = {
                os.path.basename(self._script_path): self._script_path
            }
            input_file_dict["files_to_copy"].update(files_to_copy_dict)
            self.executable = self._executable_command(
                working_directory=self.working_directory,
                script_path=self._script_path,
                enable_mpi4py=self._enable_mpi4py,
                cores=self.server.cores,
            )
            if self._enable_mpi4py:
                self.executable._mpi = True
        return input_file_dict

    def run_if_lib(self) -> None:
        """
        Compatibility function - but library run mode is not available
        """
        raise NotImplementedError(
            "Library run mode is not implemented for script jobs."
        )

    def run_static(self) -> None:
        """
        The run_static() function is called internally in pyiron to trigger the execution of the executable. This is
        typically divided into three steps: (1) the generation of the calculate function and its inputs, (2) the
        execution of this function and (3) storing the output of this function in the HDF5 file.

        In future the execution of the calculate function might be transferred to a separate process, so the separation
        in these three distinct steps is necessary to simplify the submission to an external executor.
        """
        super().run_static()
        # Update masterid for all jobs created in the working directory of the script job
        for job in self.project.iter_jobs(recursive=False, convert_to_object=False):
            pr_job = self.project.open(
                os.path.relpath(job.working_directory, self.project.path)
            )
            for subjob_id in pr_job.get_job_ids(recursive=False):
                if pr_job.db.get_item_by_id(subjob_id)["masterid"] is None:
                    pr_job.db.item_update({"masterid": str(job.job_id)}, subjob_id)

    def save_output(
        self, output_dict: Optional[dict] = None, shell_output: Optional[str] = None
    ) -> None:
        pass

    def set_input_to_read_only(self) -> None:
        """
        This function enforces read-only mode for the input classes, but it has to be implement in the individual
        classes.
        """
        super().set_input_to_read_only()
        self.input.read_only = True

    def _to_dict(self) -> dict:
        """
        Convert the job object to a dictionary.

        Returns:
            dict: A dictionary representation of the job object.
        """
        job_dict = super()._to_dict()
        job_dict["input/path"] = self._script_path
        job_dict["input/parallel"] = self._enable_mpi4py
        job_dict["input/custom_dict"] = self.input.to_builtin()
        return job_dict

    def validate_ready_to_run(self) -> None:
        """
        Validates if the job is ready to run by checking if the script path is provided.

        Raises:
            TypeError: If the script path is not provided (i.e., is None).
        """
        if self.script_path is None:
            raise TypeError(
                "ScriptJob.script_path expects a path but got None. Please provide a path before "
                + "running."
            )

    def _executable_activate_mpi(self) -> None:
        """
        Internal helper function to switch the executable to MPI mode
        """
        pass

    @staticmethod
    def _executable_command(
        working_directory: str,
        script_path: str,
        enable_mpi4py: bool = False,
        cores: int = 1,
    ) -> str:
        """
        internal function to generate the executable command to either use jupyter or python

        Args:
            working_directory (str): working directory of the current job
            script_path (str): path to the script which should be executed in the working directory
            enable_mpi4py (bool): flag to enable mpi4py
            cores (int): number of cores to use

        Returns:
            str: executable command
        """
        file_name = os.path.basename(script_path)
        path = os.path.join(working_directory, file_name)
        if file_name[-6:] == ".ipynb":
            return (
                "jupyter nbconvert --ExecutePreprocessor.timeout=9999999 --to notebook --execute "
                + path
            )
        elif file_name[-3:] == ".py" and not enable_mpi4py:
            return "python " + path
        elif file_name[-3:] == ".py" and enable_mpi4py:
            return ["mpirun", "-np", str(cores), "python", path]
        else:
            raise ValueError("Filename not recognized: ", path)



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.


def fix_ipython_autocomplete(enable: bool = True) -> None:
    """Change autocomplete behavior for IPython > 6.x

    Parameters:
        enable (bool): Whether to use the trick. Default is True.

    Notes:
    - Since IPython > 6.x, the `jedi` package is used for autocomplete by default.
    - In some cases, the autocomplete doesn't work correctly (see e.g. `here <https://github.com/ipython/ipython/issues/11653>`_).
    - To set the correct behavior, we should use the following in IPython environment:
        %config Completer.use_jedi = False
    - Alternatively, you can add the following to IPython config (`<HOME>\.ipython\profile_default\ipython_config.py`):
        c.Completer.use_jedi = False
    """

    try:
        __IPYTHON__
    except NameError:
        pass
    else:
        from IPython import __version__

        major = int(__version__.split(".")[0])
        if major >= 6:
            from IPython import get_ipython

            get_ipython().Completer.use_jedi = not enable



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
Server object class which is connected to each job containing the technical details how the job is executed.
"""

import numbers
import socket
from concurrent.futures import Executor, Future
from dataclasses import asdict, fields
from typing import Optional, Union

import pandas
from pyiron_dataclasses.v1.job import Server as ServerDataClass
from pyiron_snippets.deprecate import deprecate

from pyiron_base.interfaces.has_dict import HasDict
from pyiron_base.interfaces.lockable import Lockable, sentinel
from pyiron_base.jobs.job.extension.server.runmode import Runmode
from pyiron_base.state import state
from pyiron_base.utils.instance import static_isinstance

__author__ = "Jan Janssen"
__copyright__ = (
    "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "production"
__date__ = "Sep 1, 2017"


class Server(
    Lockable, HasDict
):  # add the option to return the job id and the hold id to the server object
    """
    Generic Server object to handle the execution environment for the job

    Args:
        host (str): hostname of the local machine
        queue (str): queue name of the currently selected queue
        cores (int): number of cores
        run_mode (pyiron_base.server.runmode.Runmode): mode of the job ['modal', 'non_modal', 'queue', 'manual']
        new_hdf (bool): create a new HDF5 file [True/False] - default=True
        accept_crash (bool): ignore execution errors raised by external executables - default False
        run_time (int): run time limit in seconds for the job to finish - required for HPC job schedulers
        memory_limit (str): memory required
        qid (int): Queuing system ID - ID received from the HPC job scheduler
        additional_arguments (dict): Additional arguments for the HPC job scheduler
        conda_environment_name (str): Name of the conda environment
        conda_environment_path (str): Path to the conda environment

    Attributes:

        .. attribute:: host

            the hostname of the current system.

        .. attribute:: queue

            the queue selected for a current simulation.

        .. attribute:: cores

            the number of cores selected for the current simulation.

        .. attribute:: run_time

            the run time in seconds selected for the current simulation.

        .. attribute:: run_mode

            the run mode of the job ['modal', 'non_modal', 'queue', 'manual']

        .. attribute:: memory_limit

            the maximum amount of RAM allocated for the calculation in GB

        .. attribute:: new_hdf

            defines whether a subjob should be stored in the same HDF5 file or in a new one.

        .. attribute:: executor

            the executor can be used to execute the job object

        .. attribute:: future

            the concurrent.futures.Future object for monitoring the execution of the job object
    """

    def __init__(
        self,
        host: Optional[str] = None,
        queue: Optional[str] = None,
        cores: int = 1,
        threads: int = 1,
        gpus: Optional[int] = None,
        run_mode: str = "modal",
        new_hdf: bool = True,
        accept_crash: bool = False,
        run_time: Optional[int] = None,
        memory_limit: Optional[str] = None,
        qid: Optional[int] = None,
        additional_arguments: dict = {},
        conda_environment_name: Optional[str] = None,
        conda_environment_path: Optional[str] = None,
    ):
        super().__init__()
        self._data = ServerDataClass(
            user=state.settings.login_user,
            host=self._init_host(host=host),
            run_mode=run_mode,
            cores=cores,
            gpus=gpus,
            threads=threads,
            new_hdf=new_hdf,
            accept_crash=accept_crash,
            run_time=run_time,
            memory_limit=memory_limit,
            queue=queue,
            qid=qid,
            additional_arguments=additional_arguments,
            conda_environment_name=conda_environment_name,
            conda_environment_path=conda_environment_path,
        )
        self._run_mode = Runmode()
        self._executor: Union[Executor, None] = None
        self._future: Union[Future, None] = None

        self.queue = queue
        self.run_mode = run_mode

    @property
    def accept_crash(self) -> bool:
        return self._data.accept_crash

    @accept_crash.setter
    @sentinel
    def accept_crash(self, accept: bool) -> None:
        self._data.accept_crash = accept

    @property
    def additional_arguments(self) -> dict:
        return self._data.additional_arguments

    @additional_arguments.setter
    @sentinel
    def additional_arguments(self, additional_arguments: dict) -> None:
        self._data.additional_arguments = additional_arguments

    @property
    def queue(self) -> Union[str, None]:
        """
        The que selected for a current simulation

        Returns:
            (str): schedulers_name
        """
        return self._data.queue

    @queue.setter
    @sentinel
    def queue(self, new_scheduler: Union[str, None]) -> None:
        """
        Set a queue for the current simulation, by choosing one of the options
        listed in :attribute:`~.queue_list`.

        Args:
            new_scheduler (str/None): scheduler name, None resets to default
                                      run_mode modal
        """
        if new_scheduler is None:
            self._data.queue = None
            self.run_mode.modal = True
            self.cores = 1
            self.threads = 1
            self._data.run_time = None
            self.memory_limit = None
        else:
            if state.queue_adapter is not None:
                (
                    cores,
                    run_time_max,
                    memory_max,
                ) = state.queue_adapter.check_queue_parameters(
                    queue=new_scheduler,
                    cores=self.cores,
                    run_time_max=self.run_time,
                    memory_max=self.memory_limit,
                )
                if self.cores is not None and cores != self.cores:
                    self._data.cores = cores
                    state.logger.warning(f"Updated the number of cores to: {cores}")
                if self.run_time is not None and run_time_max != self.run_time:
                    self._data.run_time = run_time_max
                    state.logger.warning(
                        f"Updated the run time limit to: {run_time_max}"
                    )
                if self.memory_limit is not None and memory_max != self.memory_limit:
                    self._data.memory_limit = memory_max
                    state.logger.warning(f"Updated the memory limit to: {memory_max}")
                self._data.queue = new_scheduler
                self.run_mode = "queue"
            else:
                raise TypeError(
                    "No queue adapter defined. Have you "
                    "configured in $PYIRONRESOURCES_PATHS/queues?"
                )

    @property
    def queue_id(self) -> int:
        """
        Get the queue ID - the ID in the queuing system is most likely not the same as the job ID.

        Returns:
            int: queue ID
        """
        return self._data.qid

    @queue_id.setter
    @sentinel
    def queue_id(self, qid: int) -> None:
        """
        Set the queue ID

        Args:
            qid (int): queue ID
        """
        self._data.qid = int(qid)

    @property
    def threads(self) -> int:
        """
        The number of threads selected for the current simulation

        Returns:
            (int): number of threads
        """
        return self._data.threads

    @threads.setter
    @sentinel
    def threads(self, number_of_threads: int) -> None:
        """
        The number of threads selected for the current simulation

        Args:
            number_of_threads (int): number of threads
        """
        self._data.threads = number_of_threads

    @property
    def gpus(self) -> Union[int, None]:
        """
        Total number of GPUs to use for this calculation.

        Returns:
            int: Total number of GPUs to use for this calculation.
        """
        return self._data.gpus

    @gpus.setter
    @sentinel
    def gpus(self, number_of_gpus: int) -> None:
        """
        Total number of GPUs to use for this calculation.

        Args:
            number_of_gpus (int): Total number of GPUs to use for this calculation.
        """
        self._data.gpus = number_of_gpus

    @property
    def cores(self) -> int:
        """
        The number of cores selected for the current simulation

        Returns:
            (int): number of cores
        """
        return self._data.cores

    @cores.setter
    @sentinel
    def cores(self, new_cores: int) -> None:
        """
        The number of cores selected for the current simulation

        Args:
            new_cores (int): number of cores
        """
        if not isinstance(new_cores, numbers.Integral):
            converted = int(new_cores)
            # if the conversion truncated the number, raise error otherwise silently accept it
            if new_cores != converted:
                raise ValueError(f"cores must be an integer number, not {new_cores}!")
            new_cores = converted
        if state.queue_adapter is not None and self._data.queue is not None:
            cores = state.queue_adapter.check_queue_parameters(
                queue=self.queue,
                cores=new_cores,
                run_time_max=self.run_time,
                memory_max=self.memory_limit,
            )[0]
            if cores != new_cores:
                self._data.cores = cores
                state.logger.warning(f"Updated the number of cores to: {cores}")
            else:
                self._data.cores = new_cores
        else:
            self._data.cores = new_cores

    @property
    def run_time(self) -> int:
        """
        The run time in seconds selected for the current simulation

        Returns:
            (int): run time in seconds
        """
        return self._data.run_time

    @run_time.setter
    @sentinel
    def run_time(self, new_run_time: int) -> None:
        """
        The run time in seconds selected for the current simulation

        Args:
            new_run_time (int): run time in seconds

        Raises:
            ValueError: if new_run_time cannot be converted to int
        """
        new_run_time = int(new_run_time)
        if state.queue_adapter is not None and self._data.queue is not None:
            run_time_max = state.queue_adapter.check_queue_parameters(
                queue=self.queue,
                cores=self.cores,
                run_time_max=new_run_time,
                memory_max=self.memory_limit,
            )[1]
            if run_time_max != new_run_time:
                self._data.run_time = run_time_max
                state.logger.warning(f"Updated the run time limit to: {run_time_max}")
            else:
                self._data.run_time = new_run_time
        else:
            self._data.run_time = new_run_time

    @property
    def memory_limit(self) -> str:
        return self._data.memory_limit

    @memory_limit.setter
    @sentinel
    def memory_limit(self, limit: str) -> None:
        if state.queue_adapter is not None and self._data.queue is not None:
            memory_max = state.queue_adapter.check_queue_parameters(
                queue=self.queue,
                cores=self.cores,
                run_time_max=self.run_time,
                memory_max=limit,
            )[2]
            if memory_max != limit:
                self._data.memory_limit = memory_max
                state.logger.warning(f"Updated the memory limit to: {memory_max}")
            else:
                self._data.memory_limit = limit
        else:
            self._data.memory_limit = limit

    @property
    def run_mode(self) -> str:
        """
        Get the run mode of the job

        Returns:
            (str/pyiron_base.server.runmode.Runmode): ['modal', 'non_modal', 'queue', 'manual', 'thread', 'worker',
                        'interactive', 'interactive_non_modal', 'srun', 'executor']
        """
        return self._run_mode

    @run_mode.setter
    @sentinel
    def run_mode(self, new_mode: str) -> None:
        """
        Set the run mode of the job

        Args:
            new_mode (str): ['modal', 'non_modal', 'queue', 'manual', 'thread', 'worker', 'interactive',
                        'interactive_non_modal', 'srun', 'executor']
        """
        self._run_mode.mode = new_mode
        if new_mode == "queue":
            if state.queue_adapter is None:
                raise TypeError("No queue adapter defined.")
            if self._data.queue is None:
                self.queue = state.queue_adapter.config["queue_primary"]

    @property
    def new_hdf(self) -> bool:
        """
        New_hdf5 defines whether a subjob should be stored in the same HDF5 file or in a new one.

        Returns:
            (bool): [True / False]

        """
        return self._data.new_hdf

    @new_hdf.setter
    @sentinel
    def new_hdf(self, new_hdf_bool: bool) -> None:
        """
        New_hdf5 defines whether a subjob should be stored in the same HDF5 file or in a new one.

        Args:
            new_hdf_bool (bool): [True / False]
        """
        if isinstance(new_hdf_bool, bool):
            self._data.new_hdf = new_hdf_bool
        else:
            raise TypeError(
                "The new_hdf5 is a boolean property, defining whether subjobs are stored in the same file."
            )

    @property
    def queue_list(self) -> list:
        """
        List the available Job scheduler provided by the system.

        Returns:
            (list)
        """
        return self.list_queues()

    @property
    def queue_view(self) -> pandas.DataFrame:
        """
        List the available Job scheduler provided by the system.

        Returns:
            (pandas.DataFrame)
        """
        return self.view_queues()

    @property
    def executor(self) -> Union[Executor, None]:
        """
        Executor to execute the job object this server object is attached to in the background.

        Returns:
            concurrent.futures.Executor:
        """
        if not self.run_mode.executor and self._executor is not None:
            self._executor = None
        return self._executor

    @executor.setter
    @sentinel
    def executor(self, exe: Union[Executor, None]):
        """
        Executor to execute the job object this server object is attached to in the background.

        Args:
            exe (concurrent.futures.Executor):
        """
        if isinstance(exe, Executor):
            self.run_mode.executor = True
        elif static_isinstance(exe, "flux.job.executor.FluxExecutor"):
            self.run_mode.executor = True
        elif exe is None and self.run_mode.executor:
            self.run_mode.modal = True
        elif exe is not None:
            raise TypeError(
                "The executor has to be derived from the concurrent.futures.Executor class."
            )
        self._executor = exe

    @property
    def future(self) -> Union[Future, None]:
        """
        Python concurrent.futures.Future object to track the status of the execution of the job this server object is
        attached to. This is an internal pyiron feature and most users never have to interact with the future object
        directly.

        Returns:
             concurrent.futures.Future: future object to track the status of the execution
        """
        return self._future

    # We don't wrap future in sentinel, to allow it later to be dropped to
    # None, once execution is finished
    @future.setter
    def future(self, future_obj: Future) -> None:
        """
        Set a python concurrent.futures.Future object to track the status of the execution of the job this server object
        is attached to. This is an internal pyiron feature and most users never have to interact with the future object
        directly.

        Args:
            future_obj (concurrent.futures.Future): future object to track the status of the execution
        """
        self._future = future_obj

    @property
    def conda_environment_name(self) -> str:
        """
        Get name of the conda environment the job should be executed in.

        Returns:
            str: name of the conda environment
        """
        return self._data.conda_environment_name

    @conda_environment_name.setter
    @sentinel
    def conda_environment_name(self, environment_name: str) -> None:
        """
        Set name of the conda environment the job should be executed in.

        Args:
            environment_name (str): name of the conda environment
        """
        self._data.conda_environment_name = environment_name

    @property
    def conda_environment_path(self) -> str:
        """
        Get path of the conda environment the job should be executed in.

        Returns:
            str: path of the conda environment
        """
        return self._data.conda_environment_path

    @conda_environment_path.setter
    @sentinel
    def conda_environment_path(self, environment_path: str) -> None:
        """
        Set path of the conda environment the job should be executed in.

        Args:
            environment_path (str): path of the conda environment
        """
        self._data.conda_environment_path = environment_path

    @staticmethod
    def list_queues() -> list:
        """
        List the available Job scheduler provided by the system.

        Returns:
            (list)
        """
        if state.queue_adapter is not None:
            return state.queue_adapter.queue_list
        else:
            return None

    @staticmethod
    def view_queues() -> pandas.DataFrame:
        """
        List the available Job scheduler provided by the system.

        Returns:
            (pandas.DataFrame)
        """
        if state.queue_adapter is not None:
            return state.queue_adapter.queue_view
        else:
            return None

    def _to_dict(self) -> dict:
        """
        Convert the Server object to a dictionary.

        Returns:
            dict: The Server object as a dictionary.
        """
        self._data.run_mode = self._run_mode.mode
        return asdict(self._data)

    def _from_dict(self, obj_dict: dict, version: Optional[str] = None) -> None:
        """
        Load the Server object from a dictionary.

        Args:
            obj_dict (dict): The dictionary containing the Server object data.
            version (str, optional): The version of the Server object. Defaults to None.
        """
        # backwards compatibility
        if "new_h5" in obj_dict.keys():
            obj_dict["new_hdf"] = obj_dict.pop("new_h5") == 1
        for key in ["conda_environment_name", "conda_environment_path", "qid"]:
            if key not in obj_dict.keys():
                obj_dict[key] = None
        if "accept_crash" not in obj_dict.keys():
            obj_dict["accept_crash"] = False
        if "additional_arguments" not in obj_dict.keys():
            obj_dict["additional_arguments"] = {}

        # Reload dataclass and discard unknown keys
        server_fields = tuple(f.name for f in fields(ServerDataClass))
        # force tuple otherwise dict complains about changing size
        for key in tuple(obj_dict):
            if key not in server_fields:
                del obj_dict[key]
        self._data = ServerDataClass(**obj_dict)
        self._run_mode = Runmode(mode=self._data.run_mode)

    @deprecate(message="Use job.server.to_dict() instead of to_hdf()", version=0.9)
    def to_hdf(
        self,
        hdf: "pyiron_base.storage.hdfio.ProjectHDFio",
        group_name: Optional[str] = None,
    ) -> None:
        """
        Store Server object in HDF5 file
        Args:
            hdf: HDF5 object
            group_name (str): node name in the HDF5 file
        """
        if group_name is not None:
            with hdf.open(group_name) as hdf_group:
                hdf_group["server"] = self.to_dict()
        else:
            hdf["server"] = self.to_dict()

    @deprecate(message="Use job.server.from_dict() instead of from_hdf()", version=0.9)
    def from_hdf(
        self,
        hdf: "pyiron_base.storage.hdfio.ProjectHDFio",
        group_name: Optional[str] = None,
    ) -> None:
        """
        Recover Server object in HDF5 file
        Args:
            hdf: HDF5 object
            group_name: node name in the HDF5 file
        """
        if group_name is not None:
            with hdf.open(group_name) as hdf_group:
                self.from_dict(obj_dict=hdf_group["server"])
        else:
            self.from_dict(obj_dict=hdf["server"])

    def db_entry(self) -> str:
        """
        connect all the info regarding the server into a single word that can be used e.g. as entry in a database

        Returns:
            (str): server info as single word

        """
        if self.run_mode.queue:
            server_lst = [self._data.host, str(self.cores), self.queue]
        else:
            server_lst = [self._data.host, str(self.cores)]
        return self._data.user + "@" + "#".join(server_lst)

    def __del__(self) -> None:
        """
        Delete the Server object from memory
        """
        del self._run_mode
        del self._data

    @staticmethod
    def _init_host(host) -> Union[str, None]:
        if host is None:
            return socket.gethostname()
        else:
            return host



import io
import os
import posixpath
import tarfile
import tempfile
from shutil import copytree
from typing import Tuple

import numpy as np
import pandas

from pyiron_base.state import state
from pyiron_base.utils.instance import static_isinstance


def update_id_lst(record_lst: list, job_id_lst: list) -> list:
    """
    Update the list of master IDs based on the record list and job ID list.

    Args:
        record_lst (list): List of master IDs.
        job_id_lst (list): List of job IDs.

    Returns:
        list: Updated list of master IDs.
    """
    masterid_lst = []
    for masterid in record_lst:
        if masterid is None or np.isnan(masterid):
            masterid_lst.append(None)
        elif isinstance(masterid, int) or isinstance(masterid, float):
            masterid = int(masterid)
            masterid_lst.append(job_id_lst[masterid])
    return masterid_lst


def import_jobs(
    project_instance: "pyiron_base.project.generic.Project", archive_directory: str
):
    """
    Import jobs from an archive directory to a pyiron project.

    Args:
        project_instance (pyiron_base.project.generic.Project): Pyiron project instance.
        archive_directory (str): Path to the archive directory.
    """
    # Copy HDF5 files
    # if the archive_directory is a path(string)/name of the compressed file
    if static_isinstance(
        obj=archive_directory.__class__,
        obj_type=[
            "pyiron_base.project.generic.Project",
        ],
    ):
        archive_directory = archive_directory.path
    elif not isinstance(archive_directory, str):
        raise RuntimeError(
            "The given path for importing from, does not have the correct"
            " format paths as string or pyiron Project objects are expected"
        )
    if archive_directory.endswith(".tar.gz"):
        with tempfile.TemporaryDirectory() as temp_dir:
            with tarfile.open(archive_directory, "r:gz") as tar:
                tar.extractall(path=temp_dir)
            df, common_path = transfer_files(
                origin_path=temp_dir, project_path=project_instance.path
            )
    else:
        df, common_path = transfer_files(
            origin_path=archive_directory, project_path=project_instance.path
        )

    pr_import = project_instance.open(os.curdir)
    df["project"] = [
        posixpath.normpath(
            posixpath.join(pr_import.project_path, posixpath.relpath(p, common_path))
        )
        + "/"
        for p in df["project"].values
    ]
    df["projectpath"] = len(df) * [pr_import.root_path]
    # Add jobs to database
    job_id_lst = []

    for entry in df.dropna(axis=1).to_dict(orient="records"):
        for tag in ["id", "parentid", "masterid"]:
            if tag in entry:
                del entry[tag]
        if "timestart" in entry:
            entry["timestart"] = pandas.to_datetime(entry["timestart"])
        if "timestop" in entry:
            entry["timestop"] = pandas.to_datetime(entry["timestop"])
        if "username" not in entry:
            entry["username"] = state.settings.login_user
        job_id = pr_import.db.add_item_dict(par_dict=entry, check_duplicates=True)
        job_id_lst.append(job_id)

    # Update parent and master ids
    for job_id, masterid, parentid in zip(
        job_id_lst,
        update_id_lst(record_lst=df["masterid"].values, job_id_lst=job_id_lst),
        update_id_lst(record_lst=df["parentid"].values, job_id_lst=job_id_lst),
    ):
        if masterid is not None or parentid is not None:
            pr_import.db.item_update(
                item_id=job_id, par_dict={"parentid": parentid, "masterid": masterid}
            )


def transfer_files(origin_path: str, project_path: str) -> Tuple[pandas.DataFrame, str]:
    """
    Transfer files from the origin path to the project path.

    Args:
        origin_path (str): Path to the origin directory.
        project_path (str): Path to the project directory.

    Returns:
        Tuple[pandas.DataFrame, str]: A tuple containing the job table and the common path.
    """
    df = get_dataframe(origin_path=origin_path)
    common_path = posixpath.commonpath(list(df["project"]))
    copytree(posixpath.join(origin_path, common_path), project_path, dirs_exist_ok=True)
    return df, common_path


def get_dataframe(
    origin_path: str, csv_file_name: str = "export.csv"
) -> pandas.DataFrame:
    """
    Get the job table from the csv file.

    Args:
        origin_path (str): Path to the origin directory.
        csv_file_name (str): Name of the csv file.

    Returns:
        pandas.DataFrame: Job table.
    """
    # This line looks for the csv file outside of the archive directory to
    # guarantee backward compatibility with old archives.
    if os.path.exists(csv_file_name):
        return pandas.read_csv(csv_file_name, index_col=0)
    for root, dirs, files in os.walk(origin_path):
        if csv_file_name in files:
            return pandas.read_csv(os.path.join(root, csv_file_name), index_col=0)
    raise FileNotFoundError(f"File: {csv_file_name} was not found.")


def inspect_csv(tar_path: str, csv_file: str = "export.csv") -> None:
    """
    Inspect the csv file inside a tar archive.

    Args:
        tar_path (str): Path to the tar archive.
        csv_file (str): Name of the csv file.

    Returns:
        pandas.DataFrame: Job table.
    """
    with tarfile.open(tar_path, mode="r:gz") as tar:
        for member in tar.getmembers():
            # Check if the member is a file and ends with the desired csv file name
            if member.isfile() and member.name.endswith(f"/{csv_file}"):
                # Extract the file object
                extracted_file = tar.extractfile(member)

                if extracted_file:
                    # Read the file content
                    return pandas.read_csv(
                        io.StringIO(extracted_file.read().decode("utf-8")), index_col=0
                    )
        raise FileNotFoundError(f"File: {csv_file} in {tar_path} was not found.")



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
Jobtype class to create GenericJob type objects
"""

import importlib
import inspect
import os
from typing import Optional, Union

from pyiron_snippets.singleton import Singleton

from pyiron_base.interfaces.factory import PyironFactory
from pyiron_base.jobs.job.extension.jobstatus import job_status_finished_lst
from pyiron_base.jobs.job.util import _get_safe_job_name
from pyiron_base.storage.hdfio import ProjectHDFio

__author__ = "Joerg Neugebauer, Jan Janssen"
__copyright__ = (
    "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "production"
__date__ = "Sep 1, 2017"


JOB_CLASS_DICT = {
    "FlexibleMaster": "pyiron_base.jobs.master.flexible",
    "ScriptJob": "pyiron_base.jobs.script",
    "TableJob": "pyiron_base.jobs.datamining",
    "WorkerJob": "pyiron_base.jobs.worker",
    "ExecutableContainerJob": "pyiron_base.jobs.flex.executablecontainer",
    "PythonFunctionContainerJob": "pyiron_base.jobs.flex.pythonfunctioncontainer",
}


class JobType:
    """
    The JobTypeBase class creates a new object of a given class type.
    """

    _job_class_dict = JOB_CLASS_DICT

    def __new__(
        cls,
        class_name: Union[type, str],
        project: ProjectHDFio,
        job_name: str,
        job_class_dict: Optional[dict] = None,
        delete_existing_job: bool = False,
        delete_aborted_job: bool = False,
    ):
        """
        The __new__() method allows to create objects from other classes - the class selected by class_name

        Args:
            class_name (str/Type('GenericJob')): The specific class name of the class this object belongs to.
            project (Project): Project object (defines path where job will be created and stored)
            job_name (str): name of the job (must be unique within this project path)
            job_class_dict (dict): dictionary with the jobtypes to choose from.
            delete_existing_job (bool): delete an existing job - default false
            delete_aborted_job (bool): delete an existing and aborted job - default false

        Returns:
            GenericJob: object of type class_name
        """
        if not isinstance(delete_existing_job, bool):
            raise ValueError(
                f"We got delete_existing_job = {delete_existing_job}. If you"
                " meant to delete the job, set delete_existing_job = True"
            )
        job_name = _get_safe_job_name(job_name)
        cls.job_class_dict = job_class_dict or cls._job_class_dict
        if isinstance(class_name, str):
            job_class = cls.convert_str_to_class(
                job_class_dict=cls.job_class_dict, class_name=class_name
            )
        elif inspect.isclass(class_name):
            if (
                class_name.__name__ not in cls._job_class_dict
                or cls._convert_pyiron_to_pyiron_atomistics_module(
                    class_name.__module__
                )
                == cls._job_class_dict[class_name.__name__]
            ):
                job_class = class_name
            else:
                raise ValueError(
                    f"You are trying to instantiate a job with the class {class_name} named {class_name.__name__} "
                    f"from module {class_name.__module__}, however, this name has already been registered with the"
                    f" module {cls._job_class_dict[class_name.__name__]}."
                )
        else:
            raise TypeError()
        job = job_class(project, job_name)
        if job.job_id is not None and not os.path.exists(job.project_hdf5.file_name):
            job.logger.warning(
                "No HDF5 file found - remove database entry and create new job! {}".format(
                    job.job_name
                )
            )
            delete_existing_job = True
        if delete_existing_job or (job.status.aborted and delete_aborted_job):
            job.remove()
            job = job_class(project, job_name)
        if job.status.aborted:
            job.logger.warning(
                "Job aborted - please remove it and run again! {}".format(job.job_name)
            )
        if not job.status.initialized:
            job.from_hdf()
        if job.status.string in job_status_finished_lst:
            job.set_input_to_read_only()
        return job

    @classmethod
    def unregister(cls, job_name_or_class: Union[str, type]) -> Optional[type]:
        """Unregister job type from the exposed list of available job types

        Args:
            job_name_or_class(str/type): name of the job or job class

        Returns:
            None if a str is provided as job_name_or_class
            job_name_or_class if a class is provided. Therefore, this method can be used as class decorator.
        """
        _cls = None
        if isinstance(job_name_or_class, type):
            _cls = job_name_or_class
            job_name_or_class = job_name_or_class.__name__
        if job_name_or_class in cls._job_class_dict:
            del cls._job_class_dict[job_name_or_class]
        else:
            raise KeyError(f"No JobType with name '{job_name_or_class}' found.")
        return _cls

    @staticmethod
    def _convert_pyiron_to_pyiron_atomistics_module(cls_module_str: str) -> str:
        if cls_module_str.startswith("pyiron."):
            # Currently, we set all sub-modules of pyiron_atomistics to be sub-modules of pyiron. Thus, any class
            # pyiron.submodule.PyironClass is identical to pyiron_atomistics_submodule.PyironClass:
            cls_module_str = cls_module_str.replace("pyiron.", "pyiron_atomistics.")
        return cls_module_str

    @classmethod
    def register(
        cls,
        job_class_or_module_str: Union[type, str],
        job_name: Optional[str] = None,
        overwrite: bool = False,
    ) -> None:
        """Register job type from the exposed list of available job types

        Args:
            job_class_or_module_str(type/str/None): job class itself, string representation of the job class module as
                provided by cls.__module__, or None for do not register.
            job_name(str/None): Name of the job to register. Must match cls.__name__. Can be omitted for class input.
            overwrite(bool): If True, overwrite existing job type.
        """
        if job_class_or_module_str is None:
            return
        elif isinstance(job_class_or_module_str, type):
            cls_module_str = cls._convert_pyiron_to_pyiron_atomistics_module(
                job_class_or_module_str.__module__
            )
            if job_name is not None and job_class_or_module_str.__name__ != job_name:
                raise NotImplementedError(
                    "Currently, the given name has to match the class name."
                )
            else:
                job_name = job_class_or_module_str.__name__
        elif job_name is not None:
            cls_module_str = cls._convert_pyiron_to_pyiron_atomistics_module(
                job_class_or_module_str
            )
        else:
            raise ValueError(
                "The job_name needs to be provided if a job_module_string is provided."
            )

        if (
            not overwrite
            and job_name in cls._job_class_dict
            and cls_module_str != cls._job_class_dict[job_name]
        ):
            ValueError(
                f"A JobType with name '{job_name}' is already defined! New class = '{cls_module_str}', "
                f"already registered class = '{cls._job_class_dict[job_name]}'."
            )
        else:
            cls._job_class_dict[job_name] = cls_module_str

    @staticmethod
    def convert_str_to_class(job_class_dict: dict, class_name: str) -> type:
        """
        convert the name of a class to the corresponding class object - only for pyiron internal classes.

        Args:
            job_class_dict (dict):
            class_name (str):

        Returns:
            (class):
        """
        job_type_lst = class_name.split(".")
        if len(job_type_lst) > 1:
            class_name = class_name.split()[-1][1:-2]
            job_type = class_name.split(".")[-1]
        else:
            job_type = job_type_lst[-1]

        if job_type in job_class_dict.keys():
            job_class = job_class_dict[job_type]
            if isinstance(job_class, str):
                job_module = importlib.import_module(job_class)
                job_class = getattr(job_module, job_type)
            return job_class
        raise ValueError(
            "Unknown job type: ",
            class_name,
            [job for job in list(job_class_dict.keys())],
        )


class JobFactory(PyironFactory):
    """
    The job creator is used to create job objects using pr.create.job.Code() where Code can be any external code
    which is wrapped as pyiron job type.
    """

    def __init__(self, project: "pyiron_base.project.generic.Project"):
        self._job_class_dict = JOB_CLASS_DICT
        self._project = project

    def __dir__(self) -> list:
        """
        Enable autocompletion by overwriting the __dir__() function.
        """
        return list(self._job_class_dict.keys())

    def __getattr__(self, name: str) -> callable:
        if name in self._job_class_dict:

            def wrapper(job_name, delete_existing_job=False, delete_aborted_job=False):
                """
                Create one of the following jobs:
                - 'ExampleJob': example job just generating random number
                - 'ParallelMaster': series of jobs run in parallel
                - 'ScriptJob': Python script or jupyter notebook job container
                - 'ListMaster': list of jobs

                Args:
                    job_name (str): name of the job
                    delete_existing_job (bool): delete an existing job - default false
                    delete_aborted_job (bool): delete an existing and aborted job - default false

                Returns:
                    GenericJob: job object depending on the job_type selected
                """
                job_name = _get_safe_job_name(job_name)
                return JobType(
                    class_name=name,
                    project=ProjectHDFio(
                        project=self._project.copy(), file_name=job_name
                    ),
                    job_name=job_name,
                    job_class_dict=self._job_class_dict,
                    delete_existing_job=delete_existing_job,
                    delete_aborted_job=delete_aborted_job,
                )

            return wrapper
        else:
            raise AttributeError("no job class named '{}' defined".format(name))


class JobTypeChoice(metaclass=Singleton):
    """
    Helper class to choose the job type directly from the project, autocompletion is enabled by overwriting the
    __dir__() function. This class is only required for pr.job_type.Code which is only used in pr.create_job().
    As a consequence this class can be removed once the pr.create_job() function is replaced by pr.create.job.Code().
    """

    def __init__(self):
        self._job_class_dict = None
        self.job_class_dict = JOB_CLASS_DICT

    @property
    def job_class_dict(self) -> dict:
        return self._job_class_dict

    @job_class_dict.setter
    def job_class_dict(self, job_class_dict: dict) -> None:
        self._job_class_dict = job_class_dict

    def __getattr__(self, name: str) -> str:
        if name in self._job_class_dict.keys():
            return name
        else:
            raise AttributeError("no job class named '{}' defined".format(name))

    def __dir__(self) -> list:
        """
        Enable autocompletion by overwriting the __dir__() function.
        """
        return list(self.job_class_dict.keys())



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

"""
Utility functions used in pyiron.
In order to be accessible from anywhere in pyiron, they *must* remain free of
any imports from pyiron!
"""

__author__ = "Joerg Neugebauer, Jan Janssen"
__copyright__ = (
    "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "production"
__date__ = "Sep 1, 2017"


def strtobool(val: str) -> int:
    """
    Convert a string representation of truth to true (1) or false (0).

    True values are 'y', 'yes', 't', 'true', 'on', and '1'; false values
    are 'n', 'no', 'f', 'false', 'off', and '0'.  Raises ValueError if
    'val' is anything else.

    Args:
        val (str): The string representation of truth value.

    Returns:
        int: The converted truth value as an integer (1 for true, 0 for false).

    Raises:
        ValueError: If 'val' is not a valid truth value.

    """
    val = val.lower()
    if val in ("y", "yes", "t", "true", "on", "1"):
        return 1
    elif val in ("n", "no", "f", "false", "off", "0"):
        return 0
    else:
        raise ValueError("invalid truth value %r" % (val,))



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
Classes to map the Python objects to HDF5 data structures
"""

import importlib
import numbers
import os
import posixpath
import sys
from typing import Any, Dict, List, Optional, Set, Tuple, Union

import h5py
import numpy as np
import pandas
from h5io_browser import Pointer, read_nested_dict_from_hdf
from h5io_browser.base import (
    _is_ragged_in_1st_dim_only,
    _open_hdf,
    _read_hdf,
    _write_hdf5_with_json_support,
)
from pyiron_snippets.deprecate import deprecate

from pyiron_base.interfaces.has_groups import HasGroups
from pyiron_base.jobs.job.util import _get_safe_job_name
from pyiron_base.state import state
from pyiron_base.utils.instance import static_isinstance

__author__ = "Joerg Neugebauer, Jan Janssen"
__copyright__ = (
    "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "production"
__date__ = "Sep 1, 2017"


# for historic reasons we write str(class) into the HDF 'TYPE' field of objects, so we need to parse this back out
def _extract_fully_qualified_name(type_field: str) -> str:
    """
    Extracts the fully qualified name from the given type field.

    Args:
        type_field (str): The type field containing the fully qualified name.

    Returns:
        str: The extracted fully qualified name.
    """
    return type_field.split("'")[1]


def _extract_module_class_name(type_field: str) -> Tuple[str, str]:
    """
    Extracts the module path and class name from the given type field.

    Args:
        type_field (str): The type field containing the fully qualified name.

    Returns:
        Tuple[str, str]: The module path and class name.
    """
    fully_qualified_path = _extract_fully_qualified_name(type_field)
    return fully_qualified_path.rsplit(".", maxsplit=1)


def _list_groups_and_nodes(hdf: h5py.File, h5_path: str) -> Tuple[List[str], List[str]]:
    """
    Get the list of groups and list of nodes from an open HDF5 file

    Args:
        hdf (h5py.File): file handle of an open HDF5 file
        h5_path (str): path inside the HDF5 file

    Returns:
        Tuple[List[str], List[str]]: list of groups and list of nodes
    """
    groups = set()
    nodes = set()
    try:
        h = hdf[h5_path]
        for k in h.keys():
            if isinstance(h[k], h5py.Group):
                groups.add(k)
            else:
                nodes.add(k)
    except KeyError:
        pass
    return list(groups), list(nodes)


def _import_class(module_path: str, class_name: str) -> type:
    """
    Import given class from fully qualified name and return class object.

    Args:
        module_path (str): fully qualified name of a pyiron class
        class_name (str): fully qualified name of a pyiron class

    Returns:
        type: class object of the given name
    """
    # ugly dynamic import, but only needed to log the warning anyway
    from pyiron_base.jobs.job.jobtype import JobTypeChoice

    job_class_dict = JobTypeChoice().job_class_dict  # access global singleton
    if class_name in job_class_dict:
        known_module_path = job_class_dict[class_name]
        # entries in the job_class_dict are either strings of modules or fully
        # loaded class object; in the latter case our work here is done we just
        # return the class
        if isinstance(known_module_path, type):
            return known_module_path
        if module_path != known_module_path:
            state.logger.info(
                f'Using registered module "{known_module_path}" instead of custom/old module "{module_path}" to'
                f' import job type "{class_name}"!'
            )
            module_path = known_module_path
    try:
        return getattr(
            importlib.import_module(module_path),
            class_name,
        )
    except ImportError:
        import pyiron_base.project.maintenance

        if module_path in pyiron_base.project.maintenance._MODULE_CONVERSION_DICT:
            raise RuntimeError(
                f"Could not import {class_name} from {module_path}, but module path known to have changed. "
                "Call project.maintenance.local.update_hdf_types() to upgrade storage!"
            ) from None
        else:
            raise


def _to_object(
    hdf: "FileHDFio", class_name: Optional[str] = None, **kwargs: Any
) -> Any:
    """
    Load the full pyiron object from an HDF5 file

    Args:
        hdf (FileHDFio): The HDF5 file handle.
        class_name (str, optional): If the 'TYPE' node is not available in the HDF5 file, a manual object type can be
            set. Must be as reported by `str(type(obj))`.
        **kwargs: Optional parameters to override init parameters.

    Returns:
        Any: Pyiron object of the given class_name.
    """
    if "TYPE" not in hdf.list_nodes() and class_name is None:
        raise ValueError("Objects can be only recovered from hdf5 if TYPE is given")
    elif class_name is not None and class_name != hdf.get("TYPE"):
        raise ValueError(
            "Object type in hdf5-file must be identical to input parameter"
        )
    type_field = class_name or hdf.get("TYPE")
    module_path, class_name = _extract_module_class_name(type_field)
    class_object = _import_class(module_path, class_name)

    # Backwards compatibility since the format of TYPE changed
    if type_field != str(class_object):
        hdf["TYPE"] = str(class_object)

    if hasattr(class_object, "from_hdf_args"):
        init_args = class_object.from_hdf_args(hdf)
    else:
        init_args = {}

    init_args.update(kwargs)

    obj = class_object(**init_args)
    obj.from_hdf(hdf=hdf.open(".."), group_name=hdf.h5_path.split("/")[-1])
    if static_isinstance(obj=obj, obj_type="pyiron_base.jobs.job.generic.GenericJob"):
        module_name = module_path.split(".")[0]
        module = importlib.import_module(module_name)
        if hasattr(module, "Project"):
            obj.project_hdf5._project = getattr(module, "Project")(
                obj.project_hdf5.project.path
            )
    return obj


class FileHDFio(HasGroups, Pointer):
    """
    Class that provides all info to access a h5 file. This class is based on h5io.py, which allows to
    get and put a large variety of jobs to/from h5

    Implements :class:`.HasGroups`.  Groups are HDF groups in the file, nodes are HDF datasets.

    Args:
        file_name (str): absolute path of the HDF5 file
        h5_path (str): absolute path inside the h5 path - starting from the root group
        mode (str): mode : {'a', 'w', 'r', 'r+'}, default 'a'
                    See HDFStore docstring or tables.open_file for info about modes

    .. attribute:: file_name
        absolute path to the HDF5 file
    .. attribute:: h5_path
        path inside the HDF5 file - also stored as absolute path
    .. attribute:: history
        previously opened groups / folders
    .. attribute:: file_exists
        boolean if the HDF5 was already written
    .. attribute:: base_name
        name of the HDF5 file but without any file extension
    .. attribute:: file_path
        directory where the HDF5 file is located
    .. attribute:: is_root
        boolean if the HDF5 object is located at the root level of the HDF5 file
    .. attribute:: is_open
        boolean if the HDF5 file is currently opened - if an active file handler exists
    .. attribute:: is_empty
        boolean if the HDF5 file is empty
    """

    def __init__(self, file_name: str, h5_path: str = "/", mode: str = "a") -> None:
        Pointer.__init__(self=self, file_name=file_name, h5_path=h5_path)
        self.history = []
        self._filter = ["groups", "nodes", "objects"]

    # MutableMapping Impl
    def __contains__(self, item: str) -> bool:
        """
        Check if an item exists in the HDF5 file.

        Args:
            item (str): path to the data or key of the data object

        Returns:
            bool: True if the item exists, False otherwise
        """
        nodes_groups = self.list_all()
        return item in nodes_groups["nodes"] or item in nodes_groups["groups"]

    def __len__(self) -> int:
        """
        Get the number of items in the HDF5 file.

        Returns:
            int: Number of items in the HDF5 file
        """
        nodes_groups = self.list_all()
        return len(nodes_groups["nodes"]) + len(nodes_groups["groups"])

    def __iter__(self):
        """
        Iterate over the keys in the HDF5 file.

        Returns:
            iter: Iterator over the keys in the HDF5 file
        """
        return iter(self.keys())

    def __getitem__(self, item: Union[str, slice]) -> Union[Dict, List, float, int]:
        """
        Get/ read data from the HDF5 file.

        Args:
            item (str, slice): path to the data or key of the data object

        Returns:
            Union[Dict, List, float, int]: Data or data object
        """
        if isinstance(item, slice):
            if not (item.start or item.stop or item.step):
                return self.values()
            raise NotImplementedError("Implement if needed, e.g. for [:]")
        else:
            try:
                # fast path, a good amount of accesses will want to fetch a specific dataset it knows exists in the
                # file, there's therefor no point in checking whether item is a group or a node or even worse recursing
                # in case when item contains '/'.  In most cases read_hdf5 will grab the correct data straight away and
                # if not we will still check thoroughly below.  Since list_nodes()/list_groups() each open the
                # underlying file once, this reduces the number of file opens in the most-likely case from 2 to 1 (1 to
                # check whether the data is there and 1 to read it) and increases in the worst case from 1 to 2 (1 to
                # try to read it here and one more time to verify it's not a group below).
                return _read_hdf(
                    hdf_filehandle=self.file_name, h5_path=self._get_h5_path(item)
                )
            except (ValueError, OSError, RuntimeError, NotImplementedError):
                # h5io couldn't find a dataset with name item, but there still might be a group with that name, which we
                # check in the rest of the method
                pass

            item_lst = item.split("/")
            if len(item_lst) == 1 and item_lst[0] != "..":
                # if item in self.list_nodes() we would have caught it in the fast path above
                if item in self.list_groups():
                    with self.open(item) as hdf_item:
                        obj = hdf_item.copy()
                        if self._is_convertable_dtype_object_array(obj):
                            obj = self._convert_dtype_obj_array(obj)
                        return obj
                raise ValueError(
                    "Unknown item: {} {} {}".format(item, self.file_name, self.h5_path)
                )
            else:
                if (
                    item_lst[0] == ""
                ):  # item starting with '/', thus we have an absoute HDF5 path
                    item_abs_lst = os.path.normpath(item).replace("\\", "/").split("/")
                else:  # relative HDF5 path
                    # The self.h5_path is an absolute path (/h5_path/in/h5/file), however, to
                    # reach any directory super to root, we start with a
                    # relative path = ./h5_path/in/h5/file and add whatever we get as item.
                    # The normpath finally returns a path to the item which is relative to the hdf-root.
                    item_abs_lst = (
                        os.path.normpath(os.path.join("." + self.h5_path, item))
                        .replace("\\", "/")
                        .split("/")
                    )
                # print('h5_path=', self.h5_path, 'item=', item, 'item_abs_lst=', item_abs_lst)
                if item_abs_lst[0] == "." and len(item_abs_lst) == 1:
                    # Here, we are asked to return the root of the HDF5-file. The resulting self.path would be the
                    # same as the self.file_path and, thus, the path of the pyiron Project this HDF5-file belongs to:
                    return self.create_project_from_hdf5()
                elif item_abs_lst[0] == "..":
                    # Here, we are asked to return a path super to the root of the HDF5-file, a.k.a. the path of it's
                    # pyiron Project, thus we pass the relative path to the pyiron Project to handle it:
                    return self.create_project_from_hdf5()["/".join(item_abs_lst)]
                else:
                    hdf_object = self.copy()
                    hdf_object.h5_path = "/".join(item_abs_lst[:-1])
                    return hdf_object[item_abs_lst[-1]]

    # TODO: remove this function upon 1.0.0 release
    @staticmethod
    def _is_convertable_dtype_object_array(obj: np.ndarray) -> bool:
        """
        Check if an object array is convertable to a different dtype.

        Args:
            obj (np.ndarray): Object array

        Returns:
            bool: True if the object array is convertable, False otherwise
        """
        if isinstance(obj, np.ndarray) and obj.dtype == np.dtype(object):
            first_element = obj[(0,) * obj.ndim]
            last_element = obj[(-1,) * obj.ndim]
            if (
                isinstance(first_element, numbers.Number)
                and isinstance(last_element, numbers.Number)
                and not _is_ragged_in_1st_dim_only(obj)
            ):
                return True
        return False

    # TODO: remove this function upon 1.0.0 release
    @staticmethod
    def _convert_dtype_obj_array(obj: np.ndarray) -> np.ndarray:
        """
        Convert an object array to a different dtype.

        Args:
            obj (np.ndarray): Object array

        Returns:
            np.ndarray: Converted object array
        """
        try:
            result = np.array(obj.tolist())
        except ValueError:
            result = np.array(obj.tolist(), dtype=object)
        if result.dtype != np.dtype(object):
            state.logger.warning(
                f"Deprecated data structure! "
                f"Returned array was converted from dtype='O' to dtype={result.dtype} "
                f"via `np.array(result.tolist())`.\n"
                f"Please run rewrite_hdf5() (from a job: job.project_hdf5.rewrite_hdf5() ) to update this data! "
                f"To update all your data run Project.maintenance.update.base_v0_3_to_v0_4('all')."
            )
            return result
        else:
            return obj

    def __setitem__(
        self,
        key: str,
        value: Union[pandas.DataFrame, pandas.Series, Dict, List, float, int],
    ) -> None:
        """
        Store data inside the HDF5 file.

        Args:
            key (str): Key to store the data
            value (Union[pandas.DataFrame, pandas.Series, Dict, List, float, int]): Data to store
        """
        if hasattr(value, "to_hdf") & (
            not isinstance(value, (pandas.DataFrame, pandas.Series))
        ):
            value.to_hdf(self, key)
            return
        _write_hdf5_with_json_support(
            hdf_filehandle=self.file_name,
            h5_path=self._get_h5_path(key),
            data=value,
        )

    @property
    def base_name(self) -> str:
        """
        Get the name of the HDF5 file without the file extension.

        Returns:
            str: Name of the HDF5 file without the file extension
        """
        return ".".join(posixpath.basename(self.file_name).split(".")[:-1])

    @property
    def file_path(self) -> str:
        """
        Get the directory where the HDF5 file is located.

        Returns:
            str: Directory where the HDF5 file is located
        """
        return posixpath.dirname(self.file_name)

    def get_size(self, hdf: "FileHDFio") -> float:
        """
        Get the size of the groups inside the HDF5 file.

        Args:
            hdf (FileHDFio): HDF5 file

        Returns:
            float: File size in Bytes
        """
        return sum([sys.getsizeof(hdf[p]) for p in hdf.list_nodes()]) + sum(
            [self.get_size(hdf[p]) for p in hdf.list_groups()]
        )

    def copy(self) -> "FileHDFio":
        """
        Copy the Python object which links to the HDF5 file - in contrast to copy_to() which copies the content of the
        HDF5 file to a new location.

        Returns:
            FileHDFio: New FileHDFio object pointing to the same HDF5 file
        """
        new_h5 = FileHDFio(file_name=self.file_name, h5_path=self.h5_path)
        new_h5._filter = self._filter
        return new_h5

    def create_group(self, name: str, track_order: bool = False) -> "FileHDFio":
        """
        Create an HDF5 group - similar to a folder in the filesystem - the HDF5 groups allow the users to structure
        their data.

        Args:
            name (str): Name of the HDF5 group
            track_order (bool): If False, this groups tracks its elements in alphanumeric order,
                                if True, in insertion order

        Returns:
            FileHDFio: FileHDFio object pointing to the new group
        """
        full_name = self._get_h5_path(name)
        with _open_hdf(self.file_name, mode="a") as h:
            try:
                h.create_group(full_name, track_order=track_order)
            except ValueError:
                pass
        h_new = self[name].copy()
        return h_new

    def remove_group(self) -> None:
        """
        Remove an HDF5 group if it exists. If the group does not exist, no error message is raised.
        """
        try:
            with _open_hdf(self.file_name, mode="a") as hdf_file:
                del hdf_file[self.h5_path]
        except KeyError:
            pass

    def open(self, h5_rel_path: str) -> "FileHDFio":
        """
        Create an HDF5 group and enter this specific group. If the group exists in the HDF5 path,
        only the h5_path is set correspondingly, otherwise the group is created first.

        Args:
            h5_rel_path (str): Relative path from the current HDF5 path - h5_path - to the new group

        Returns:
            FileHDFio: FileHDFio object pointing to the new group
        """
        new_h5_path = self.copy()
        if os.path.isabs(h5_rel_path):
            raise ValueError(
                "Absolute paths are not supported -> replace by relative path name!"
            )

        if h5_rel_path.strip() == ".":
            h5_rel_path = ""
        if h5_rel_path.strip() != "":
            new_h5_path.h5_path = self._get_h5_path(h5_rel_path)
        new_h5_path.history.append(h5_rel_path)

        return new_h5_path

    def close(self) -> None:
        """
        Close the current HDF5 path and return to the path before the last open.
        """
        path_lst = self.h5_path.split("/")
        last = self.history[-1].strip()
        if len(last) > 0:
            hist_lst = last.split("/")
            self.h5_path = "/".join(path_lst[: -len(hist_lst)])
            if len(self.h5_path.strip()) == 0:
                self.h5_path = "/"
        del self.history[-1]

    def show_hdf(self) -> None:
        """
        Iterate over the HDF5 data structure and generate a human-readable graph.
        """
        self._walk()

    def remove_file(self) -> None:
        """
        Remove the HDF5 file with all the related content.
        """
        if self.file_exists:
            os.remove(self.file_name)

    def get_from_table(self, path: str, name: str) -> Union[Dict, List, float, int]:
        """
        Get a specific value from a pandas.DataFrame.

        Args:
            path (str): Relative path to the data object
            name (str): Parameter key

        Returns:
            Union[Dict, List, float, int]: The value associated with the specific parameter key
        """
        df_table = self.get(path)
        keys = df_table["Parameter"]
        if name in keys:
            job_id = keys.index(name)
            return df_table["Value"][job_id]
        raise ValueError("Unknown name: {0}".format(name))

    def get_pandas(self, name: str) -> pandas.DataFrame:
        """
        Load a dictionary from the HDF5 file and display the dictionary as a pandas DataFrame.

        Args:
            name (str): HDF5 node name

        Returns:
            pd.DataFrame: The dictionary as a pandas DataFrame object
        """
        val = self.get(name)
        if isinstance(val, dict):
            df = pandas.DataFrame(val)
            return df

    def get(
        self, key: str, default: Optional[object] = None
    ) -> Union[Dict, List, float, int]:
        """
        Get data from the HDF5 file.

        Args:
            key (str): Path to the data or key of the data object
            default (object): Default value to return if key doesn't exist

        Returns:
            Union[Dict, List, float, int]: Data or data object
        """
        try:
            return self.__getitem__(key)
        except ValueError:
            if default is not None:
                return default
            else:
                raise

    def put(
        self,
        key: str,
        value: Union[pandas.DataFrame, pandas.Series, Dict, List, float, int],
    ) -> None:
        """
        Store data inside the HDF5 file.

        Args:
            key (str): Key to store the data
            value (Union[pandas.DataFrame, pandas.Series, Dict, List, float, int]): Data to store
        """
        self.__setitem__(key=key, value=value)

    def _list_all(self) -> Dict[str, List[str]]:
        """
        List all groups and nodes of the HDF5 file - where groups are equivalent to directories and nodes to files.

        Returns:
            Dict[str, List[str]]: Dictionary with keys "groups" and "nodes" containing lists of groups and nodes
        """
        if self.file_exists:
            with _open_hdf(self.file_name) as hdf:
                groups, nodes = _list_groups_and_nodes(hdf=hdf, h5_path=self.h5_path)
            iopy_nodes = self._filter_io_objects(set(groups))
            return {
                "groups": sorted(list(set(groups) - iopy_nodes)),
                "nodes": sorted(list((set(nodes) - set(groups)).union(iopy_nodes))),
            }
        else:
            return {"groups": [], "nodes": []}

    def _list_nodes(self) -> List[str]:
        """
        List all nodes in the HDF5 file.

        Returns:
            List[str]: List of nodes in the HDF5 file
        """
        return self.list_all()["nodes"]

    def _list_groups(self) -> List[str]:
        """
        List all groups in the HDF5 file.

        Returns:
            List[str]: List of groups in the HDF5 file
        """
        return self.list_all()["groups"]

    def listdirs(self) -> List[str]:
        """
        Equivalent to os.listdirs (consider groups as equivalent to dirs).

        Returns:
            List[str]: List of groups in pytables for the path self.h5_path
        """
        return self.list_groups()

    def list_dirs(self) -> List[str]:
        """
        Equivalent to os.listdirs (consider groups as equivalent to dirs).

        Returns:
            List[str]: List of groups in pytables for the path self.h5_path
        """
        return self.list_groups()

    def keys(self) -> List[str]:
        """
        List all groups and nodes of the HDF5 file - where groups are equivalent to directories and nodes to files.

        Returns:
            List[str]: All groups and nodes
        """
        list_all_dict = self.list_all()
        return list_all_dict["nodes"] + list_all_dict["groups"]

    def values(self) -> List[Union[Dict, List, float, int]]:
        """
        List all values for all groups and nodes of the HDF5 file.

        Returns:
            List[Union[Dict, List, float, int]]: List of all values
        """
        return [self[key] for key in self.keys()]

    def items(self) -> List[Tuple[str, Union[Dict, List, float, int]]]:
        """
        List all keys and values as items of all groups and nodes of the HDF5 file.

        Returns:
            List[Tuple[str, Union[Dict, List, float, int]]]: List of sets (key, value)
        """
        return [(key, self[key]) for key in self.keys()]

    def groups(self) -> "FileHDFio":
        """
        Filter HDF5 file by groups.

        Returns:
            FileHDFio: An HDF5 file which is filtered by groups
        """
        new = self.copy()
        new._filter = ["groups"]
        return new

    def nodes(self) -> "FileHDFio":
        """
        Filter HDF5 file by nodes.

        Returns:
            FileHDFio: An HDF5 file which is filtered by nodes
        """
        new = self.copy()
        new._filter = ["nodes"]
        return new

    def hd_copy(
        self,
        hdf_old: "FileHDFio",
        hdf_new: "FileHDFio",
        exclude_groups: Optional[List[str]] = None,
        exclude_nodes: Optional[List[str]] = None,
    ) -> None:
        """
        Copy data from one HDF5 file to another.

        Args:
            hdf_old (FileHDFio): Source HDF5 file
            hdf_new (FileHDFio): Destination HDF5 file
            exclude_groups (List[str]): List of groups to exclude from the copy
            exclude_nodes (List[str]): List of nodes to exclude from the copy
        """
        if exclude_groups is None or len(exclude_groups) == 0:
            exclude_groups_split = list()
            group_list = hdf_old.list_groups()
        else:
            exclude_groups_split = [i.split("/", 1) for i in exclude_groups]
            check_groups = [i[-1] for i in exclude_groups_split]
            group_list = list(
                (set(hdf_old.list_groups()) ^ set(check_groups))
                & set(hdf_old.list_groups())
            )

        if exclude_nodes is None or len(exclude_nodes) == 0:
            exclude_nodes_split = list()
            node_list = hdf_old.list_nodes()
        else:
            exclude_nodes_split = [i.split("/", 1) for i in exclude_nodes]
            check_nodes = [i[-1] for i in exclude_nodes_split]
            node_list = list(
                (set(hdf_old.list_nodes()) ^ set(check_nodes))
                & set(hdf_old.list_nodes())
            )
        hdf_new.write_dict(data_dict={p: hdf_old[p] for p in node_list})
        for p in group_list:
            h_new = hdf_new.create_group(p)
            ex_n = [e[-1] for e in exclude_nodes_split if p == e[0] or len(e) == 1]
            ex_g = [e[-1] for e in exclude_groups_split if p == e[0] or len(e) == 1]
            self.hd_copy(hdf_old[p], h_new, exclude_nodes=ex_n, exclude_groups=ex_g)
        return hdf_new

    @deprecate(job_name="ignored!", exclude_groups="ignored!", exclude_nodes="ignored!")
    def rewrite_hdf5(
        self,
        job_name: Optional[str] = None,
        info: bool = False,
        exclude_groups: Optional[List[str]] = None,
        exclude_nodes: Optional[List[str]] = None,
    ) -> None:
        """
        Rewrite the entire hdf file.

        Args:
            job_name (Optional[str]): Deprecated argument, ignored.
            info (bool): Whether to give the information on how much space has been saved.
            exclude_groups (Optional[List[str]]): List of groups to exclude from the copy.
            exclude_nodes (Optional[List[str]]): List of nodes to exclude from the copy.
        """
        if job_name is not None:
            state.logger.warning(
                "Specifying job_name is deprecated and ignored! Future versions will change signature."
            )
        file_name = self.file_name
        new_file = file_name + "_rewrite"

        self_hdf = FileHDFio(file_name=file_name)
        hdf_new = FileHDFio(file_name=new_file, h5_path="/")

        old_logger_level = state.logger.level
        state.logger.level = 50
        hdf_new = self.hd_copy(self_hdf, hdf_new)
        state.logger.level = old_logger_level

        if info:
            print(
                "compression rate from old to new: {}".format(
                    self.file_size(self_hdf) / self.file_size(hdf_new)
                )
            )
            print(
                "data size vs file size: {}".format(
                    self.get_size(hdf_new) / self.file_size(hdf_new)
                )
            )
        self.remove_file()
        os.rename(hdf_new.file_name, file_name)

    def __str__(self) -> str:
        """
        Machine readable string representation

        Returns:
            str: list all nodes and groups as string
        """
        return self.__repr__()

    def __repr__(self) -> str:
        """
        Human readable string representation

        Returns:
            str: list all nodes and groups as string
        """
        return str(self.list_all())

    def __del__(self):
        del self._file_name
        del self.history
        del self._h5_path

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Compatibility function for the with statement
        """
        self.close()
        try:
            self._store.close()
        except AttributeError:
            pass

    def _read(self, item: str) -> Union[Dict, List, float, int]:
        """
        Internal read function to read data from the HDF5 file

        Args:
            item (str): path to the data or key of the data object

        Returns:
            dict, list, float, int: data or data object
        """
        return _read_hdf(hdf_filehandle=self.file_name, h5_path=self._get_h5_path(item))

    def write_dict_to_hdf(self, data_dict: dict) -> None:
        """
        Write a dictionary to HDF5

        Args:
            data_dict (dict): dictionary with objects which should be written to HDF5
        """
        self.write_dict(data_dict=data_dict)

    def read_dict_from_hdf(
        self, group_paths: List[str] = [], recursive: bool = False
    ) -> dict:
        """
        Read data from HDF5 file into a dictionary - by default only the nodes are converted to dictionaries, additional
        sub groups can be specified using the group_paths parameter.

        Args:
            group_paths (List[str]): list of additional groups to be included in the dictionary, for example:
                                     ["input", "output", "output/generic"]
                                     These groups are defined relative to the h5_path.
            recursive (bool): Load all subgroups recursively

        Returns:
            Dict: The loaded data. Can be of any type supported by ``write_hdf5``.
        """
        return read_nested_dict_from_hdf(
            file_name=self.file_name,
            h5_path=self.h5_path,
            group_paths=group_paths,
            recursive=recursive,
            slash="ignore",
        )

    def create_project_from_hdf5(self) -> "Project":
        """
        Internal function to create a pyiron project pointing to the directory where the HDF5 file is located.

        Returns:
            Project: pyiron project object
        """
        from pyiron_base.project.generic import Project

        return Project(path=self.file_path)

    def _get_h5_path(self, name: str) -> str:
        """
        Internal function to combine the current h5_path with the relative path

        Args:
            name (str): relative path

        Returns:
            str: combined path
        """
        return posixpath.join(self.h5_path, name)

    def _get_h5io_type(self, name: str) -> str:
        """
        Internal function to get h5io type

        Args:
            name (str): HDF5 key

        Returns:
            str: h5io type
        """
        with _open_hdf(self.file_name) as store:
            return str(store[self.h5_path][name].attrs.get("TITLE", ""))

    def _filter_io_objects(self, groups: Union[List[str], Set[str]]) -> Set[str]:
        """
        Internal function to extract h5io objects (which have the same type as normal groups)

        Args:
            groups (list, set): list of groups (as obtained e.g. from listdirs

        Returns:
            set: h5io objects
        """
        h5io_types = (
            "dict",
            "list",
            "tuple",
            "pd_dataframe",
            "pd_series",
            "multiarray",
            "json",
        )
        group_h5io = set(
            [group for group in groups if self._get_h5io_type(group) in h5io_types]
        )
        return group_h5io

    def _walk(self, level: int = 0) -> None:
        """
        Internal helper function for show_hdf() - iterating over the HDF5 datastructure and generating a human readable
        graph.

        Args:
            level (int): iteration level
        """
        l_dict = self.list_all()
        indent = level * "  "
        for node in l_dict["nodes"]:
            print(indent + "node", node)
        for group in l_dict["groups"]:
            print(indent + "group: ", group)
            with self.open(group) as hdf_group:
                hdf_group._walk(level=level + 1)


class BaseHDFio:
    """
    Dummy class to allow other code to type check if it received a ProjectHDFio
    or DummyHDFio object.  Usually this is used to check if it is safe to call
    `to_object` on this object.
    """

    pass


class ProjectHDFio(FileHDFio, BaseHDFio):
    """
    The ProjectHDFio class connects the FileHDFio and the Project class, it is derived from the FileHDFio class but in
    addition the a project object instance is located at self.project enabling direct access to the database and other
    project related functionality, some of which are mapped to the ProjectHDFio class as well.

    Args:
        project (Project): pyiron Project the current HDF5 project is located in
        file_name (str): name of the HDF5 file - in contrast to the FileHDFio object where file_name represents the
                         absolute path of the HDF5 file.
        h5_path (str): absolute path inside the h5 path - starting from the root group
        mode (str): mode : {'a', 'w', 'r', 'r+'}, default 'a'
                    See HDFStore docstring or tables.open_file for info about modes

    Attributes:

        .. attribute:: project

            Project instance the ProjectHDFio object is located in

        .. attribute:: root_path

            the pyiron user directory, defined in the .pyiron configuration

        .. attribute:: project_path

            the relative path of the current project / folder starting from the root path
            of the pyiron user directory

        .. attribute:: path

            the absolute path of the current project / folder plus the absolute path in the HDF5 file as one path

        .. attribute:: file_name

            absolute path to the HDF5 file

        .. attribute:: h5_path

            path inside the HDF5 file - also stored as absolute path

        .. attribute:: history

            previously opened groups / folders

        .. attribute:: file_exists

            boolean if the HDF5 was already written

        .. attribute:: base_name

            name of the HDF5 file but without any file extension

        .. attribute:: file_path

            directory where the HDF5 file is located

        .. attribute:: is_root

            boolean if the HDF5 object is located at the root level of the HDF5 file

        .. attribute:: is_open

            boolean if the HDF5 file is currently opened - if an active file handler exists

        .. attribute:: is_empty

            boolean if the HDF5 file is empty

        .. attribute:: user

            current unix/linux/windows user who is running pyiron

        .. attribute:: sql_query

            an SQL query to limit the jobs within the project to a subset which matches the SQL query.

        .. attribute:: db

            connection to the SQL database

        .. attribute:: working_directory

            working directory of the job is executed in - outside the HDF5 file
    """

    def __init__(
        self,
        project: "pyiron_base.project.generic.Project",
        file_name: str,
        h5_path: Optional[str] = None,
        mode: Optional[str] = None,
    ) -> None:
        self._file_name = _get_safe_filename(file_name)
        if h5_path is None:
            h5_path = "/"
        self._project = project.copy()
        super(ProjectHDFio, self).__init__(
            file_name=os.path.join(self._project.path, self._file_name).replace(
                "\\", "/"
            ),
            h5_path=h5_path,
            mode=mode,
        )

    @property
    def base_name(self) -> str:
        """
        The absolute path to of the current pyiron project - absolute path on the file system, not including the HDF5
        path.

        Returns:
            str: current project path
        """
        return self._project.path

    @property
    def db(self) -> "DatabaseAccess":
        """
        Get connection to the SQL database

        Returns:
            DatabaseAccess: database conncetion
        """
        return self._project.db

    @property
    def path(self) -> str:
        """
        Absolute path of the HDF5 group starting from the system root - combination of the absolute system path plus the
        absolute path inside the HDF5 file starting from the root group.

        Returns:
            str: absolute path
        """
        return os.path.join(self._project.path, self.h5_path[1:]).replace("\\", "/")

    @property
    def project(self) -> "pyiron_base.project.generic.Project":
        """
        Get the project instance the ProjectHDFio object is located in

        Returns:
            Project: pyiron project
        """
        return self._project

    @property
    def project_path(self) -> str:
        """
        the relative path of the current project / folder starting from the root path
        of the pyiron user directory

        Returns:
            str: relative path of the current project / folder
        """
        return self._project.project_path

    @property
    def root_path(self) -> str:
        """
        the pyiron user directory, defined in the .pyiron configuration

        Returns:
            str: pyiron user directory of the current project
        """
        return self._project.root_path

    @property
    def sql_query(self) -> str:
        """
        Get the SQL query for the project

        Returns:
            str: SQL query
        """
        return self._project.sql_query

    @sql_query.setter
    def sql_query(self, new_query: str) -> None:
        """
        Set the SQL query for the project

        Args:
            new_query (str): SQL query
        """
        self._project.sql_query = new_query

    @property
    def user(self) -> str:
        """
        Get current unix/linux/windows user who is running pyiron

        Returns:
            str: username
        """
        return self._project.user

    @property
    def working_directory(self) -> str:
        """
        Get the working directory of the current ProjectHDFio object. The working directory equals the path but it is
        represented by the filesystem:
            /absolute/path/to/the/file.h5/path/inside/the/hdf5/file
        becomes:
            /absolute/path/to/the/file_hdf5/path/inside/the/hdf5/file

        Returns:
            str: absolute path to the working directory
        """
        project_full_path = "/".join(self.file_name.split("/")[:-1])
        file_name = self.file_name.split("/")[-1]
        if ".h5" in file_name:
            file_name = file_name.split(".h5")[0]
        file_name += "_hdf5"
        if self.h5_path[0] == "/":
            h5_path = self.h5_path[1:]
        else:
            h5_path = self.h5_path
        return posixpath.join(project_full_path, file_name, h5_path)

    @property
    def _filter(self) -> str:
        """
        Get project filter

        Returns:
            str: project filter
        """
        return self._project._filter

    @_filter.setter
    def _filter(self, new_filter: str) -> None:
        """
        Set project filter

        Args:
            new_filter (str): project filter
        """
        self._project._filter = new_filter

    @property
    def _inspect_mode(self) -> bool:
        """
        Check if inspect mode is activated

        Returns:
            bool: [True/False]
        """
        return self._project._inspect_mode

    @_inspect_mode.setter
    def _inspect_mode(self, read_mode: bool) -> None:
        """
        Activate or deactivate inspect mode

        Args:
            read_mode (bool): [True/False]
        """
        self._project._inspect_mode = read_mode

    @property
    def name(self) -> str:
        """
        Get the name of the HDF5 group.

        Returns:
            str: The name of the HDF5 group.
        """
        return os.path.basename(self.h5_path)

    def copy(self) -> "ProjectHDFio":
        """
        Copy the ProjectHDFio object - copying just the Python object but maintaining the same pyiron path

        Returns:
            ProjectHDFio: copy of the ProjectHDFio object
        """
        new_h5 = ProjectHDFio(
            project=self._project, file_name=self._file_name, h5_path=self._h5_path
        )
        new_h5._filter = self._filter
        return new_h5

    def create_hdf(self, path: str, job_name: str) -> "ProjectHDFio":
        """
        Create an ProjectHDFio object to store project related information - for testing aggregated data

        Args:
            path (str): absolute path
            job_name (str): name of the HDF5 container

        Returns:
            ProjectHDFio: HDF5 object
        """
        return self._project.create_hdf(path=path, job_name=job_name)

    def create_working_directory(self) -> None:
        """
        Create the working directory on the file system if it does not exist already.
        """
        os.makedirs(self.working_directory, exist_ok=True)

    def to_object(self, class_name: Optional[str] = None, **kwargs) -> object:
        """
        Load the full pyiron object from an HDF5 file

        Args:
            class_name(str, optional): if the 'TYPE' node is not available in
                        the HDF5 file a manual object type can be set,
                        must be as reported by `str(type(obj))`
            **kwargs: optional parameters optional parameters to override init
                      parameters

        Returns:
            pyiron object of the given class_name
        """
        return _to_object(self, class_name, **kwargs)

    def get_job_id(self, job_specifier: Union[str, int]) -> int:
        """
        get the job_id for job named job_name in the local project path from database

        Args:
            job_specifier (str, int): name of the job or job ID

        Returns:
            int: job ID of the job
        """
        return self._project.get_job_id(job_specifier=job_specifier)

    def inspect(self, job_specifier: Union[str, int]) -> "JobCore":
        """
        Inspect an existing pyiron object - most commonly a job - from the database

        Args:
            job_specifier (str, int): name of the job or job ID

        Returns:
            JobCore: Access to the HDF5 object - not a GenericJob object - use load() instead.
        """
        return self._project.inspect(job_specifier=job_specifier)

    def load(
        self, job_specifier: Union[str, int], convert_to_object: bool = True
    ) -> Union["GenericJob", "JobCore"]:
        """
        Load an existing pyiron object - most commonly a job - from the database

        Args:
            job_specifier (str, int): name of the job or job ID
            convert_to_object (bool): convert the object to an pyiron object or only access the HDF5 file - default=True
                                      accessing only the HDF5 file is about an order of magnitude faster, but only
                                      provides limited functionality. Compare the GenericJob object to JobCore object.

        Returns:
            GenericJob, JobCore: Either the full GenericJob object or just a reduced JobCore object
        """
        return self._project.load(
            job_specifier=job_specifier, convert_to_object=convert_to_object
        )

    def load_from_jobpath(
        self,
        job_id: Optional[int] = None,
        db_entry: Optional[dict] = None,
        convert_to_object: bool = True,
    ) -> Union["GenericJob", "JobCore"]:
        """
        Internal function to load an existing job either based on the job ID or based on the database entry dictionary.

        Args:
            job_id (int, optional): Job ID - optional, but either the job_id or the db_entry is required.
            db_entry (dict, optional): database entry dictionary - optional, but either the job_id or the db_entry is required.
            convert_to_object (bool): convert the object to an pyiron object or only access the HDF5 file - default=True
                                      accessing only the HDF5 file is about an order of magnitude faster, but only
                                      provides limited functionality. Compare the GenericJob object to JobCore object.

        Returns:
            GenericJob, JobCore: Either the full GenericJob object or just a reduced JobCore object
        """
        return self._project.load_from_jobpath(
            job_id=job_id, db_entry=db_entry, convert_to_object=convert_to_object
        )

    def remove_job(
        self, job_specifier: Union[str, int], _unprotect: bool = False
    ) -> None:
        """
        Remove a single job from the project based on its job_specifier.

        Args:
            job_specifier (Union[str, int]): Name of the job or job ID.
            _unprotect (bool): [True/False] Delete the job without validating the dependencies to other jobs.
                               Default is False.
        """
        self._project.remove_job(job_specifier=job_specifier, _unprotect=_unprotect)

    def create_project_from_hdf5(self) -> "Project":
        """
        Internal function to create a pyiron project pointing to the directory where the HDF5 file is located.

        Returns:
            Project: pyiron project object
        """
        return self._project.__class__(path=self.file_path)


class DummyHDFio(HasGroups, BaseHDFio):
    """
    A dummy ProjectHDFio implementation to serialize objects into a dict
    instead of a HDF5 file.

    It is modeled after ProjectHDFio, but supports just enough methods to
    successfully write objects.

    After all desired objects have been written to it, you may extract a pure
    dict from with with `.to_dict`.

    A simple example for storing data containers:

    >>> from pyiron_base import DataContainer, Project
    >>> pr = Project(...)
    >>> hdf = DummyHDFio(pr, '/', {})
    >>> d = DataContainer({'a': 42, 'b':{'c':4, 'g':33}})
    >>> d.to_hdf(hdf)
    >>> hdf.to_dict()
    {'READ_ONLY': False,
     'a__index_0': 42,
     'b__index_1': {
         'READ_ONLY': False,
         'c__index_0': 4,
         'g__index_1': 33,
         'NAME': 'DataContainer',
         'TYPE': "<class
         'pyiron_base.storage.datacontainer.DataContainer'>",
         'OBJECT': 'DataContainer',
         'VERSION': '0.1.0',
         'HDF_VERSION': '0.2.0'
     },
     'NAME': 'DataContainer',
     'TYPE': "<class
     'pyiron_base.storage.datacontainer.DataContainer'>",
     'OBJECT': 'DataContainer',
     'VERSION': '0.1.0',
     'HDF_VERSION': '0.2.0'}
    """

    def __init__(
        self,
        project,
        h5_path: str,
        cont: Optional[dict] = None,
        root: Optional["DummyHDFio"] = None,
    ):
        """

        Args:
            project (Project): the project this object should advertise itself
                               belong to; in practice it is not often used for
                               writing objects
            h5_path (str): the path of the HDF group this object fakes
            cont (dict, optional): dict to save written values into, make a new
                                   one if not given
            root (DummyHDFio, optional): if this object will be a child of
                                         another one, the parent must be passed
                                         here, to make hdf['..'] work.
        """
        self._project = project
        self._dict = {}
        self._h5_path = h5_path
        self._root = root
        if cont is not None:
            self.write_dict_to_hdf(cont)

    def __getitem__(self, item: str) -> Union["DummyHDFio", Any]:
        """
        Return a value from storage.

        If `item` is in :meth:`.list_groups()` this must return another :class:`.GenericStorage`.

        Args:
            item (str): name of value

        Returns:
            :class:`.GenericStorage`: if `item` refers to a sub group
            object: value that is stored under `item`

        Raises:
            ValueError: `item` is neither a node or a sub group of this group
        """
        try:
            v = self._dict[item]
            if isinstance(v, DummyHDFio) and v._empty():
                raise KeyError()
            else:
                return v
        except KeyError:
            if item == "..":
                return self._root
            # compat with ProjectHDFio with for some reasons raises ValueErrors
            raise ValueError(item) from None

    def get(
        self, key: Union[str, slice], default: Optional[object] = None
    ) -> Union[dict, list, float, int]:
        """
        Internal wrapper function for __getitem__() - self[name]

        Args:
            key (str, slice): path to the data or key of the data object
            default (object): default value to return if key doesn't exist

        Returns:
            dict, list, float, int: data or data object
        """
        try:
            return self[key]
        except ValueError:
            if default is not None:
                return default
            else:
                raise

    def __setitem__(self, item: str, value: Any) -> None:
        self._dict[item] = value

    def create_group(self, name: str) -> "DummyHDFio":
        """
        Create a new sub group.

        Args:
            name (str): name of the new group
        """
        if name == "..":
            return self._root
        d = self._dict.get(name, None)
        if d is None:
            self._dict[name] = d = type(self)(
                self._project, os.path.join(self.h5_path, name), cont={}, root=self
            )
        elif isinstance(d, DummyHDFio):
            pass
        else:
            raise RuntimeError(f"'{name}' is already a node!")
        return d

    def _list_nodes(self) -> List[str]:
        return [k for k, v in self._dict.items() if not isinstance(v, DummyHDFio)]

    def _list_groups(self) -> List[str]:
        return [
            k
            for k, v in self._dict.items()
            if isinstance(v, DummyHDFio) and not v._empty()
        ]

    def __contains__(self, item) -> bool:
        return item in self._dict

    @property
    def project(self) -> "Project":
        if self._project is not None:
            return self._project
        else:
            raise RuntimeError("No project set!")

    @property
    def h5_path(self) -> str:
        return self._h5_path

    def open(self, name: str) -> "DummyHDFio":
        """
        Descend into a sub group.

        If `name` does not exist yet, create a new group.  Calling :meth:`~.close` on the returned object returns this
        object.

        Args:
            name (str): name of sub group

        Returns:
            :class:`.GenericStorage`: sub group
        """
        # FIXME: what if name in self.list_nodes()
        new = self.create_group(name)
        new._prev = self
        return new

    def close(self) -> "DummyHDFio":
        """
        Surface from a sub group.

        If this object was not returned from a previous call to :meth:`.open` it returns itself silently.
        """
        try:
            return self._prev
        except AttributeError:
            return self

    def __enter__(self) -> "DummyHDFio":
        """
        Compatibility function for the with statement
        """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Compatibility function for the with statement
        """
        self.close()

    def to_dict(self) -> dict:
        """
        Convert the HDF5 data to a dictionary.

        Returns:
            dict: The converted dictionary.
        """

        def unwrap(v):
            if isinstance(v, DummyHDFio):
                return v.to_dict()
            return v

        return {k: unwrap(v) for k, v in self._dict.items()}

    def to_object(self, class_name: Optional[str] = None, **kwargs) -> object:
        """
        Load the full pyiron object from an HDF5 file

        Args:
            class_name(str, optional): if the 'TYPE' node is not available in
                        the HDF5 file a manual object type can be set,
                        must be as reported by `str(type(obj))`
            **kwargs: optional parameters optional parameters to override init
                      parameters

        Returns:
            pyiron object of the given class_name
        """
        return _to_object(self, class_name, **kwargs)

    def _empty(self) -> bool:
        """
        Check if the DummyHDFio object is empty.

        Returns:
            bool: True if the object is empty, False otherwise.
        """
        if len(self._dict) == 0:
            return True
        return len(self.list_nodes()) == 0 and all(
            self[g]._empty() for g in self.list_groups()
        )

    def write_dict_to_hdf(self, data_dict: dict) -> None:
        """
        Write a dictionary to the HDF5 file.

        Args:
            data_dict (dict): The dictionary to be written to the HDF5 file.
        """
        for k, v in data_dict.items():
            if isinstance(v, dict):
                g = self.create_group(k)
                g.write_dict_to_hdf(v)
            else:
                self[k] = v

    def read_dict_from_hdf(
        self, group_paths: List[str] = [], recursive: bool = False
    ) -> Union[dict, Any]:
        """
        Read data from the HDF5 file and return it as a dictionary.

        Args:
            group_paths (List[str]): List of group paths to read data from.
            recursive (bool): If True, read data recursively from all groups.

        Returns:
            Union[dict, Any]: The read data as a dictionary or any other object if recursive is True.
        """
        if recursive:
            return self.to_dict()

        data = {}
        for path in group_paths:
            keys = path.split("/")
            try:
                d = self[key]
                for key in keys[1:]:
                    d = d[key]
            except KeyError:
                d = None
            if isinstance(d, DummyHDFio):
                d = d.to_object()
            data[path] = d
        return data


def _get_safe_filename(file_name: str) -> str:
    """
    Get a safe filename by replacing special characters and adding a file extension.

    Args:
        file_name (str): The original file name.

    Returns:
        str: The safe file name with a file extension.
    """
    file_path_no_ext, file_ext = os.path.splitext(file_name)
    file_path = os.path.dirname(file_path_no_ext)
    file_name_no_ext = os.path.basename(file_path_no_ext)
    file_name = os.path.join(
        file_path, _get_safe_job_name(name=file_name_no_ext) + file_ext
    )
    file_name += ".h5" if not file_name.endswith(".h5") else ""
    return file_name.replace("\\", "/")



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

# API of the pyiron_base module - in alphabetical order
from pyiron_base.interfaces.factory import PyironFactory
from pyiron_base.interfaces.has_groups import HasGroups
from pyiron_base.interfaces.has_hdf import HasHDF
from pyiron_base.interfaces.object import HasDatabase, HasStorage, PyironObject
from pyiron_base.jobs.datamining import PyironTable, TableJob
from pyiron_base.jobs.dynamic import warn_dynamic_job_classes
from pyiron_base.jobs.flex.factory import create_job_factory
from pyiron_base.jobs.job.extension.executable import Executable
from pyiron_base.jobs.job.extension.jobstatus import (
    JobStatus,
    job_status_finished_lst,
    job_status_lst,
    job_status_successful_lst,
)
from pyiron_base.jobs.job.extension.server.queuestatus import validate_que_request
from pyiron_base.jobs.job.factory import JobFactoryCore
from pyiron_base.jobs.job.generic import GenericJob
from pyiron_base.jobs.job.interactive import InteractiveBase
from pyiron_base.jobs.job.jobtype import JOB_CLASS_DICT, JobType, JobTypeChoice
from pyiron_base.jobs.job.template import PythonTemplateJob, TemplateJob
from pyiron_base.jobs.job.toolkit import BaseTools, Toolkit
from pyiron_base.jobs.master.flexible import FlexibleMaster
from pyiron_base.jobs.master.generic import GenericMaster, get_function_from_string
from pyiron_base.jobs.master.interactivewrapper import InteractiveWrapper
from pyiron_base.jobs.master.list import ListMaster
from pyiron_base.jobs.master.parallel import JobGenerator, ParallelMaster
from pyiron_base.project.decorator import job
from pyiron_base.project.external import Notebook, dump, load
from pyiron_base.project.generic import Creator, Project
from pyiron_base.state import state
from pyiron_base.state.install import install_dialog
from pyiron_base.state.settings import Settings
from pyiron_base.storage.datacontainer import DataContainer
from pyiron_base.storage.flattenedstorage import FlattenedStorage
from pyiron_base.storage.has_stored_traits import HasStoredTraits
from pyiron_base.storage.hdfio import FileHDFio, ProjectHDFio
from pyiron_base.storage.inputlist import InputList
from pyiron_base.storage.parameters import GenericParameters
from pyiron_base.utils.jedi import fix_ipython_autocomplete
from pyiron_base.utils.parser import Logstatus, extract_data_from_file

# Internal init
from ._version import get_versions

Project.register_tools("base", BaseTools)

# Set version of pyiron_base
__version__ = get_versions()["version"]

# Jedi fix
fix_ipython_autocomplete()

# Dynamic job class definition is no longer supported in pyiron_base >=0.7.0
warn_dynamic_job_classes(
    resource_folder_lst=state.settings.resource_paths,
    logger=state.logger,
)

__all__ = [
    PyironFactory,
    FlattenedStorage,
    FileHDFio,
    ProjectHDFio,
    DataContainer,
    HasStoredTraits,
    InputList,
    GenericParameters,
    Executable,
    Notebook,
    load,
    dump,
    create_job_factory,
    validate_que_request,
    GenericJob,
    InteractiveBase,
    JobStatus,
    job_status_successful_lst,
    job_status_finished_lst,
    job_status_lst,
    JOB_CLASS_DICT,
    JobType,
    JobTypeChoice,
    TemplateJob,
    PythonTemplateJob,
    JobFactoryCore,
    FlexibleMaster,
    GenericMaster,
    get_function_from_string,
    InteractiveWrapper,
    ListMaster,
    ParallelMaster,
    JobGenerator,
    Creator,
    Logstatus,
    extract_data_from_file,
    Settings,
    install_dialog,
    PyironTable,
    TableJob,
    HasDatabase,
    HasStorage,
    PyironObject,
    HasGroups,
    HasHDF,
    Toolkit,
]



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

"""Classes to help developers avoid code duplication when writing tests for pyiron."""

import doctest
import os
import unittest
from abc import ABC
from contextlib import redirect_stdout
from inspect import getfile
from io import StringIO

import numpy as np

from pyiron_base import PythonTemplateJob, state
from pyiron_base.project.generic import Project

__author__ = "Liam Huber"
__copyright__ = (
    "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "0.0"
__maintainer__ = "Liam Huber"
__email__ = "huber@mpie.de"
__status__ = "development"
__date__ = "Mar 23, 2021"


class PyironTestCase(unittest.TestCase, ABC):
    """
    Base class for all pyiron unit tets.

    Registers utility type equality functions:
        - np.testing.assert_array_equal

    Optionally includes testing the docstrings in the specified module by
    overloading :attr:`~.docstring_module`.
    """

    def setUp(self):
        self.addTypeEqualityFunc(np.ndarray, self._assert_equal_numpy)

    def _assert_equal_numpy(self, a, b, msg=None):
        try:
            np.testing.assert_array_equal(a, b, err_msg=msg if msg is not None else "")
        except AssertionError as e:
            raise self.failureException(*e.args) from None

    @classmethod
    def setUpClass(cls):
        cls._initial_settings_configuration = state.settings.configuration.copy()
        if any([cls is c for c in _TO_SKIP]):
            raise unittest.SkipTest(f"{cls.__name__} tests, it's a base class")
        super().setUpClass()

    @classmethod
    def tearDownClass(cls) -> None:
        state.update(cls._initial_settings_configuration)

    @property
    def docstring_module(self):
        """
        Define module whose docstrings will be tested
        """
        return None

    def test_docstrings(self):
        """
        Fails with output if docstrings in the given module fails.

        Output capturing adapted from https://stackoverflow.com/a/22434594/12332968
        """
        with StringIO() as buf, redirect_stdout(buf):
            result = doctest.testmod(self.docstring_module)
            output = buf.getvalue()
        self.assertFalse(result.failed > 0, msg=output)


class TestWithProject(PyironTestCase, ABC):
    """
    Tests that start and remove a project for their suite.
    """

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        print("TestWithProject: Setting up test project")
        cls.project_path = getfile(cls)[:-3].replace("\\", "/")
        cls.file_location, cls.project_name = os.path.split(cls.project_path)
        cls.project = Project(cls.project_path)

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        cls.project.remove(enable=True)
        try:
            os.remove(os.path.join(cls.file_location, "pyiron.log"))
        except FileNotFoundError:
            pass


class TestWithCleanProject(TestWithProject, ABC):
    """
    Tests that start and remove a project for their suite, and remove jobs from the project for each test.
    """

    def tearDown(self):
        super().tearDown()
        self.project.remove_jobs(recursive=True, progress=False, silently=True)


class ToyJob(PythonTemplateJob):
    def __init__(self, project, job_name):
        """A toyjob to test export/import functionalities."""
        super(ToyJob, self).__init__(project, job_name)
        self.input.data_in = 100

    def write_input(self):
        super().write_input()
        self.input.write(os.path.join(self.working_directory, "input.yml"))

    # Allow writing of the input file
    def _check_if_input_should_be_written(self):
        return True

    # Check for valid input
    def validate_ready_to_run(self):
        if not isinstance(self.input.data_in, int):
            raise ValueError(
                f"data_in in should be of type int, not {type(self.input.data_in)}."
            )

    # This function is executed
    def run_static(self):
        self.status.running = True
        self.output.data_out = self.input.data_in + 1
        self.status.finished = True
        self.to_hdf()
        self.compress()


class TestWithFilledProject(TestWithProject, ABC):
    """
    Tests that creates a projects, creates jobs and sub jobs in it, and at the end of unit testing,
    removes the project.
    """

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        job = cls.project.create_job(job_type=ToyJob, job_name="toy_1")
        job.run()
        job = cls.project.create_job(job_type=ToyJob, job_name="toy_2")
        job.run()
        job.status.aborted = True

        cls.pr_sub = cls.project.open("sub_project")
        job = cls.pr_sub.create_job(job_type=ToyJob, job_name="toy_1")
        job.run()
        job = cls.pr_sub.create_job(job_type=ToyJob, job_name="toy_2")
        job.run()
        job.status.suspended = True
        job = cls.pr_sub.create_job(job_type=ToyJob, job_name="toy_3")
        job.run()

        cls.n_jobs_filled_with = 5
        # In a number of tests we compare the found jobs to an expected number of jobs
        # Let's code that number once here instead of magic-numbering it throughout
        # the tests


_TO_SKIP = [
    PyironTestCase,
    TestWithProject,
    TestWithCleanProject,
    TestWithFilledProject,
]



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
Helper functions for the JobCore and GenericJob objects
"""

import os
import posixpath
import shutil
import stat
import tarfile
from itertools import islice
from typing import Optional, Tuple, Union

import monty.io
import psutil

from pyiron_base.database.sqlcolumnlength import JOB_STR_LENGTH
from pyiron_base.utils.instance import static_isinstance
from pyiron_base.utils.safetar import safe_extract

__author__ = "Jan Janssen"
__copyright__ = (
    "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "production"
__date__ = "Nov 28, 2020"


_special_symbol_replacements = {
    ".": "d",
    "-": "m",
    "+": "p",
    ",": "c",
    " ": "_",
}


def _copy_database_entry(
    new_job_core: "pyiron_base.jobs.job.generic.GenericJob", job_copied_id: int
) -> None:
    """
    Copy database entry from previous job

    Args:
        new_job_core (GenericJob): Copy of the job object
        job_copied_id (int): Job id of the copied job
    """
    db_entry = new_job_core.project.db.get_item_by_id(job_copied_id)
    if db_entry is not None:
        db_entry["job"] = new_job_core.job_name
        db_entry["subjob"] = new_job_core.project_hdf5.h5_path
        db_entry["project"] = new_job_core.project_hdf5.project_path
        db_entry["projectpath"] = new_job_core.project_hdf5.root_path
        del db_entry["id"]
        job_id = new_job_core.project.db.add_item_dict(db_entry)
        new_job_core.reset_job_id(job_id=job_id)


def _copy_to_delete_existing(
    project_class: "pyiron_base.project.generic.Project",
    job_name: str,
    delete_job: bool,
):
    """
    Check if the job exists already in the project, if that is the case either
    delete it or reload it depending on the setting of delete_job

    Args:
        project_class (Project): The project to copy the job to.
            (Default is None, use the same project.)
        job_name (str): The new name to assign the duplicate job. Required if the project is `None` or the same
            project as the copied job. (Default is None, try to keep the same name.)
        delete_job (bool): Delete job if it exists already

    Returns:
        GenericJob/ None
    """
    job_table = project_class.job_table(recursive=False)
    if len(job_table) > 0 and job_name in job_table.job.values:
        if not delete_job:
            return project_class.load(job_name)
        else:
            project_class.remove_job(job_name)
            return None


def _get_project_for_copy(
    job: "pyiron_base.jobs.job.core.JobCore",
    project: Optional[
        Union[
            "pyiron_base.project.generic.Project",
            "pyiron_base.jobs.job.core.JobCore",
            "pyiron_base.storage.hdfio.ProjectHDFio",
        ]
    ],
    new_job_name: str,
) -> Tuple[
    "pyiron_base.project.generic.Project", "pyiron_base.storage.hdfio.ProjectHDFio"
]:
    """
    Internal helper function to generate a project and hdf5 project for copying

    Args:
        job (JobCore): Job object used for comparison
        project (JobCore/ProjectHDFio/Project/None): The project to copy the job to.
            (Default is None, use the same project.)
        new_job_name (str): The new name to assign the duplicate job. Required if the
            project is `None` or the same project as the copied job. (Default is None,
            try to keep the same name.)

    Returns:
        Project, ProjectHDFio
    """
    if static_isinstance(
        obj=project.__class__, obj_type="pyiron_base.jobs.job.core.JobCore"
    ):
        file_project = project.project
        hdf5_project = project.project_hdf5.open(new_job_name)
    elif isinstance(job.project, project.__class__):
        file_project = project
        hdf5_project = job.project_hdf5.__class__(
            project=project, file_name=new_job_name, h5_path="/" + new_job_name
        )
    elif isinstance(job.project_hdf5, project.__class__):
        file_project = project.project
        hdf5_project = project.open(new_job_name)
    elif project is None:
        file_project = job.project
        hdf5_project = job.project_hdf5.__class__(
            project=file_project, file_name=new_job_name, h5_path="/" + new_job_name
        )
    else:
        raise ValueError("Project should be JobCore/ProjectHDFio/Project/None")
    return file_project, hdf5_project


def _get_safe_job_name(
    name: Union[str, tuple],
    ndigits: Optional[int] = 8,
    special_symbols: Optional[dict] = None,
) -> str:
    """
    Sanitize a job name, optionally appending numeric values.

    Args:
        name (str|tuple): The name to sanitize, or a tuple of the name and any number
            of numeric values to append with '_' in between.
        ndigits (int|None): How many digits to round any floating point values in a
            `name` tuple to. (Default is 8; to not round at all use None.)
        special_symbols (dict|None): Conversions of special symbols to apply. This will
            be applied to the default conversion dict, which contains:
            DEFAULT_CONV_DICT

    Returns:
        (str): The sanitized (and possibly rounded) name.
    """
    d_special_symbols = _special_symbol_replacements.copy()
    if special_symbols is not None:
        d_special_symbols.update(special_symbols)

    def round_(value, ndigits=ndigits):
        if isinstance(value, float) and ndigits is not None:
            return round(value, ndigits=ndigits)
        return value

    if not isinstance(name, str):
        name_rounded = [round_(nn) for nn in name]
        job_name = "_".join([str(nn) for nn in name_rounded])
    else:
        job_name = name
    for k, v in d_special_symbols.items():
        job_name = job_name.replace(k, v)
    _is_valid_job_name(job_name=job_name)
    return job_name


_get_safe_job_name.__doc__ = _get_safe_job_name.__doc__.replace(
    "DEFAULT_CONV_DICT", f"{_special_symbol_replacements}"
)


def _rename_job(
    job: Union[
        "pyiron_base.jobs.job.generic.GenericJob", "pyiron_base.jobs.job.core.JobCore"
    ],
    new_job_name: str,
) -> None:
    """ """
    new_job_name = _get_safe_job_name(new_job_name)
    child_ids = job.child_ids
    if child_ids:
        for child_id in child_ids:
            ham = job.project.load(child_id)
            ham.move_to(job.project.open(new_job_name + "_hdf5"))
    old_working_directory = job.working_directory
    if len(job.project_hdf5.h5_path.split("/")) > 2:
        new_location = job.project_hdf5.open("../" + new_job_name)
    else:
        new_location = job.project_hdf5.__class__(
            job.project, new_job_name, h5_path="/" + new_job_name
        )
    if job.job_id:
        job.project.db.item_update(
            {"job": new_job_name, "subjob": new_location.h5_path}, job.job_id
        )
    old_job_name = job.name
    job._name = new_job_name
    job.project_hdf5.copy_to(destination=new_location, maintain_name=False)
    job.project_hdf5.remove_file()
    job.project_hdf5 = new_location
    if os.path.exists(old_working_directory):
        shutil.move(old_working_directory, job.working_directory)
        os.rmdir("/".join(old_working_directory.split("/")[:-1]))
    if os.path.exists(os.path.join(job.working_directory, old_job_name + ".tar.bz2")):
        os.rename(
            os.path.join(job.working_directory, old_job_name + ".tar.bz2"),
            os.path.join(job.working_directory, job.job_name + ".tar.bz2"),
        )


def _is_valid_job_name(job_name: str) -> None:
    """
    internal function to validate the job_name - only available in Python 3.4 <

    Args:
        job_name (str): job name
    """
    if not job_name.isidentifier():
        raise ValueError(
            f'Invalid name for a pyiron object, must be letters, digits (not as first character) and "_" only, not {job_name}'
        )
    if len(job_name) > JOB_STR_LENGTH:
        raise ValueError(
            f"Invalid name for a PyIron object: must be less then or equal to {JOB_STR_LENGTH} characters"
        )


def _get_restart_copy_dict(job: "pyiron_base.jobs.job.generic.GenericJob") -> dict:
    copy_dict = {}
    for i, actual_name in enumerate(
        [os.path.basename(f) for f in job.restart_file_list]
    ):
        if actual_name in job.restart_file_dict.keys():
            new_name = job.restart_file_dict[actual_name]
        else:
            new_name = os.path.basename(job.restart_file_list[i])
        copy_dict[new_name] = job.restart_file_list[i]
    return copy_dict


def _copy_restart_files(job: "pyiron_base.jobs.job.generic.GenericJob") -> None:
    """
    Internal helper function to copy the files required for the restart job.
    """
    if not (os.path.isdir(job.working_directory)):
        raise ValueError(
            "The working directory is not yet available to copy restart files"
        )
    for file_name, source in _get_restart_copy_dict(job=job).items():
        shutil.copy(
            source,
            posixpath.join(job.working_directory, file_name),
        )


def _kill_child(job: "pyiron_base.jobs.job.core.JobCore") -> None:
    """
    Internal helper function to kill a child process.

    Args:
        job (JobCore): job object to decompress
    """
    if (
        static_isinstance(
            obj=job.__class__, obj_type="pyiron_base.jobs.master.GenericMaster"
        )
        and not job.server.run_mode.queue
        and (job.status.running or job.status.submitted)
    ):
        for proc in psutil.process_iter():
            try:
                pinfo = proc.as_dict(attrs=["pid", "cwd"])
            except psutil.NoSuchProcess:
                pass
            else:
                if pinfo["cwd"] is not None and pinfo["cwd"].startswith(
                    job.working_directory
                ):
                    job_process = psutil.Process(pinfo["pid"])
                    job_process.kill()


def _job_compressed_name(job: "pyiron_base.jobs.job.core.JobCore") -> str:
    """Return the canonical file name of a compressed job."""
    return _get_compressed_job_name(working_directory=job.working_directory)


def _get_compressed_job_name(
    working_directory: "pyiron_base.jobs.job.core.JobCore",
) -> str:
    """Return the canonical file name of a compressed job from the working directory."""
    return os.path.join(
        working_directory, os.path.basename(working_directory) + ".tar.bz2"
    )


def _job_compress(
    job: "pyiron_base.jobs.job.core.JobCore",
    files_to_compress: list = [],
    files_to_remove: list = [],
) -> None:
    """
    Compress the output files of a job object.

    Args:
        job (JobCore): job object to compress
        files_to_compress (list): list of files to compress
        files_to_remove (list): list of files to remove
    """

    def delete_file_or_folder(fullname):
        if os.path.isfile(fullname):
            os.remove(fullname)
        elif os.path.isdir(fullname):
            shutil.rmtree(fullname)

    if not _job_is_compressed(job):
        for name in files_to_remove:
            delete_file_or_folder(fullname=os.path.join(job.working_directory, name))
        if len(files_to_compress) == 0:
            return
        cwd = os.getcwd()
        try:
            os.chdir(job.working_directory)
            with tarfile.open(_job_compressed_name(job), "w:bz2") as tar:
                for name in files_to_compress:
                    if "tar" not in name and not stat.S_ISFIFO(os.stat(name).st_mode):
                        tar.add(name)
            for name in files_to_compress:
                if name != _job_compressed_name(job):
                    delete_file_or_folder(
                        fullname=os.path.join(job.working_directory, name)
                    )
        finally:
            os.chdir(cwd)
    else:
        job.logger.info("The files are already compressed!")


def _job_decompress(job: "pyiron_base.jobs.job.core.JobCore") -> None:
    """
    Decompress the output files of a compressed job object.

    Args:
        job (JobCore): job object to decompress
    """
    tar_file_name = _job_compressed_name(job)
    try:
        with tarfile.open(tar_file_name, "r:bz2") as tar:
            safe_extract(tar, job.working_directory)
        os.remove(tar_file_name)
    except IOError:
        pass


def _working_directory_is_compressed(working_directory: str) -> bool:
    """
    Check if the working directory of a given job is already compressed or not.

    Args:
        working_directory (str): working directory of the job object

    Returns:
        bool: [True/False]
    """
    compressed_name = os.path.basename(
        _get_compressed_job_name(working_directory=working_directory)
    )
    return compressed_name in os.listdir(working_directory)


def _job_is_compressed(job: "pyiron_base.jobs.job.core.JobCore") -> bool:
    """
    Check if the job is already compressed or not.

    Args:
        job (JobCore): job object to check

    Returns:
        bool: [True/False]
    """
    return _working_directory_is_compressed(working_directory=job.working_directory)


def _working_directory_list_files(
    working_directory: str, include_archive: bool = True
) -> list:
    """
    Returns list of files in the jobs working directory.

    If the working directory is compressed, return a list of files in the archive.

    Args:
        working_directory (str): working directory of the job object to inspect files in
        include_archive (bool): include files in the .tag.gz archive

    Returns:
        list of str: file names
    """
    if os.path.isdir(working_directory):
        uncompressed_files_lst = os.listdir(working_directory)
        if include_archive and _working_directory_is_compressed(
            working_directory=working_directory
        ):
            compressed_job_name = _get_compressed_job_name(
                working_directory=working_directory
            )
            with tarfile.open(compressed_job_name, "r") as tar:
                compressed_files_lst = [
                    member.name for member in tar.getmembers() if member.isfile()
                ]
                uncompressed_files_lst.remove(os.path.basename(compressed_job_name))
                return uncompressed_files_lst + compressed_files_lst
        else:
            return uncompressed_files_lst
    return []


def _job_list_files(job: "pyiron_base.jobs.job.core.JobCore") -> list:
    """
    Returns list of files in the jobs working directory.

    If the job is compressed, return a list of files in the archive.

    Args:
        job (JobCore): job object to inspect files in

    Returns:
        list of str: file names
    """
    return _working_directory_list_files(working_directory=job.working_directory)


def _working_directory_read_file(
    working_directory: str, file_name: str, tail: Optional[int] = None
) -> list:
    """
    Return list of lines of the given file.

    Transparently decompresses the file if working directory is compressed.

    If `tail` is given and job is decompressed, only read the last lines
    instead of traversing the full file.

    Args:
        working_directory (str): working directory of the job object
        file_name (str): the file to print
        tail (int, optional): only return the last lines

    Raises:
        FileNotFoundError: if the given file name does not exist in the job folder
    """
    if file_name not in _working_directory_list_files(
        working_directory=working_directory
    ):
        raise FileNotFoundError(file_name)

    if _working_directory_is_compressed(
        working_directory=working_directory
    ) and file_name not in os.listdir(working_directory):
        with tarfile.open(
            _get_compressed_job_name(working_directory=working_directory),
            encoding="utf8",
        ) as f:
            lines = [
                line.decode("utf8") for line in f.extractfile(file_name).readlines()
            ]
            if tail is None:
                return lines
            else:
                return lines[-tail:]
    else:
        file_name = posixpath.join(working_directory, file_name)
        if tail is None:
            with open(file_name) as f:
                return f.readlines()
        else:
            lines = list(
                reversed(
                    [
                        line + os.linesep
                        for line in islice(monty.io.reverse_readfile(file_name), tail)
                    ]
                )
            )
            # compatibility with the other methods
            # monty strips all newlines, where as reading the other ways does
            # not.  So if a file does not end with a newline (as most text
            # files) adding it to every line like above adds an additional one.
            if len(lines) > 0:
                lines[-1] = lines[-1].rstrip(os.linesep)
            return lines


def _job_read_file(
    job: "pyiron_base.jobs.job.core.JobCore", file_name: str, tail: Optional[int] = None
) -> list:
    """
    Return list of lines of the given file.

    Transparently decompresses the file if job is compressed.

    If `tail` is given and job is decompressed, only read the last lines
    instead of traversing the full file.

    Args:
        file_name (str): the file to print
        tail (int, optional): only return the last lines

    Raises:
        FileNotFoundError: if the given file name does not exist in the job folder
    """
    return _working_directory_read_file(
        working_directory=job.working_directory, file_name=file_name, tail=tail
    )


def _job_archive(job: "pyiron_base.jobs.job.core.JobCore") -> None:
    """
    Compress HDF5 file of the job object to tar-archive

    Args:
        job (JobCore): job object to archive
    """
    fpath = job.project_hdf5.file_path
    jname = job.job_name
    h5_dir_name = jname + "_hdf5"
    h5_file_name = jname + ".h5"
    cwd = os.getcwd()
    try:
        os.chdir(fpath)
        with tarfile.open(
            os.path.join(fpath, job.job_name + ".tar.bz2"), "w:bz2"
        ) as tar:
            for name in [h5_dir_name, h5_file_name]:
                tar.add(name)
        for name in [h5_dir_name, h5_file_name]:
            fullname = os.path.join(fpath, name)
            if os.path.isfile(fullname):
                os.remove(fullname)
            elif os.path.isdir(fullname):
                shutil.rmtree(fullname)
    finally:
        os.chdir(cwd)


def _job_unarchive(job: "pyiron_base.jobs.job.core.JobCore") -> None:
    """
    Decompress HDF5 file of the job object from tar-archive

    Args:
        job (JobCore): job object to unarchive
    """
    fpath = job.project_hdf5.file_path
    try:
        tar_name = os.path.join(fpath, job.job_name + ".tar.bz2")
        with tarfile.open(tar_name, "r:bz2") as tar:
            safe_extract(tar, fpath)
        os.remove(tar_name)
    finally:
        pass


def _job_is_archived(job: "pyiron_base.jobs.job.core.JobCore") -> bool:
    """
    Check if the HDF5 file of the Job is compressed as tar-archive

    Args:
        job (JobCore): job object to check

    Returns:
        bool: [True/False]
    """
    return os.path.isfile(
        os.path.join(job.project_hdf5.file_path, job.job_name + ".tar.bz2")
    )


def _job_delete_hdf(job: "pyiron_base.jobs.job.core.JobCore") -> None:
    """
    Delete HDF5 file of job object

    Args:
        job (JobCore): job object to delete
    """
    if os.path.isfile(job.project_hdf5.file_name):
        os.remove(job.project_hdf5.file_name)


def _job_delete_files(job: "pyiron_base.jobs.job.core.JobCore") -> None:
    """
    Delete files in the working directory of job object

    Args:
        job (JobCore): job object to delete
    """
    working_directory = str(job.working_directory)
    if job._import_directory is None and os.path.exists(working_directory):
        shutil.rmtree(working_directory)
    else:
        job._import_directory = None


def _job_remove_folder(job: "pyiron_base.jobs.job.core.JobCore") -> None:
    """
    Delete the working directory of the job object

    Args:
        job (JobCore): job object to delete
    """
    working_directory = os.path.abspath(os.path.join(str(job.working_directory), ".."))
    if os.path.exists(working_directory) and len(os.listdir(working_directory)) == 0:
        shutil.rmtree(working_directory)


def _job_store_before_copy(job: "pyiron_base.jobs.job.core.JobCore") -> bool:
    """
    Store job in HDF5 file for copying

    Args:
        job (GenericJob): job object to copy

    Returns:
        bool: [True/False] if the HDF5 file of the job exists already
    """
    if not job.project_hdf5.file_exists:
        delete_file_after_copy = True
    else:
        delete_file_after_copy = False
    job.to_hdf()
    return delete_file_after_copy


def _job_reload_after_copy(
    job: "pyiron_base.jobs.job.core.JobCore", delete_file_after_copy: bool
) -> None:
    """
    Reload job from HDF5 file after copying

    Args:
        job (GenericJob): copied job object
        delete_file_after_copy (bool): delete HDF5 file after reload
    """
    job.from_hdf()
    if delete_file_after_copy:
        job.project_hdf5.remove_file()



import inspect
from typing import Optional

from pyiron_base.jobs.job.extension.server.generic import Server
from pyiron_base.project.generic import Project


# The combined decorator
def job(
    funct: Optional[callable] = None,
    *,
    host: Optional[str] = None,
    queue: Optional[str] = None,
    cores: int = 1,
    threads: int = 1,
    gpus: Optional[int] = None,
    run_mode: str = "modal",
    new_hdf: bool = True,
    accept_crash: bool = False,
    run_time: Optional[int] = None,
    memory_limit: Optional[str] = None,
    qid: Optional[int] = None,
    additional_arguments: dict = {},
    conda_environment_name: Optional[str] = None,
    conda_environment_path: Optional[str] = None,
    output_file_lst: list = [],
    output_key_lst: list = [],
):
    """
    Decorator to create a pyiron job object from any python function

    Args:
        funct (callable): python function to create a job object from
        host (str): the hostname of the current system.
        queue (str): the queue selected for a current simulation.
        cores (int): the number of cores selected for the current simulation.
        threads (int): the number of threads selected for the current simulation.
        gpus (int): the number of gpus selected for the current simulation.
        run_mode (str): the run mode of the job ['modal', 'non_modal', 'queue', 'manual']
        new_hdf (bool): defines whether a subjob should be stored in the same HDF5 file or in a new one.
        accept_crash (bool): ignore execution errors raised by external executables - default False
        run_time (int): run time limit in seconds for the job to finish - required for HPC job schedulers
        memory_limit (str): memory required
        qid (int): Queuing system ID - ID received from the HPC job scheduler
        additional_arguments (dict): Additional arguments for the HPC job scheduler
        conda_environment_name (str): Name of the conda environment
        conda_environment_path (str): Path to the conda environment
        output_file_lst (list):
        output_key_lst (list):

    Returns:
        callable: The decorated functions

    Example:
        >>> from pyiron_base import job, Project
        >>>
        >>> @job
        >>> def my_function_a(a, b=8):
        >>>     return a + b
        >>>
        >>> @job(cores=2)
        >>> def my_function_b(a, b=8):
        >>>     return a + b
        >>>
        >>> pr = Project("test")
        >>> c = my_function_a(a=1, b=2, pyiron_project=pr)
        >>> d = my_function_b(a=c, b=3, pyiron_project=pr)
        >>> print(d.pull())

        Output: 6
    """

    def get_delayed_object(
        *args,
        pyiron_project: Project,
        python_function: callable,
        pyiron_resource_dict: dict,
        resource_default_dict: dict,
        list_length: Optional[int] = None,
        **kwargs,
    ):
        for k, v in resource_default_dict.items():
            if k not in pyiron_resource_dict:
                pyiron_resource_dict[k] = v
        delayed_job_object = pyiron_project.wrap_python_function(
            python_function=python_function,
            *args,
            job_name=None,
            automatically_rename=True,
            execute_job=False,
            delayed=True,
            output_file_lst=output_file_lst,
            output_key_lst=output_key_lst,
            list_length=list_length,
            **kwargs,
        )
        delayed_job_object._server = Server(**pyiron_resource_dict)
        return delayed_job_object

    # This is the actual decorator function that applies to the decorated function
    def pyiron_job_function(f) -> callable:
        def function(
            *args,
            pyiron_project: Project = Project("."),
            pyiron_resource_dict: dict = {},
            list_length: Optional[int] = None,
            **kwargs,
        ):
            resource_default_dict = {
                "host": host,
                "queue": queue,
                "cores": cores,
                "threads": threads,
                "gpus": gpus,
                "run_mode": run_mode,
                "new_hdf": new_hdf,
                "accept_crash": accept_crash,
                "run_time": run_time,
                "memory_limit": memory_limit,
                "qid": qid,
                "additional_arguments": additional_arguments,
                "conda_environment_name": conda_environment_name,
                "conda_environment_path": conda_environment_path,
            }
            return get_delayed_object(
                *args,
                python_function=f,
                pyiron_project=pyiron_project,
                pyiron_resource_dict=pyiron_resource_dict,
                resource_default_dict=resource_default_dict,
                list_length=list_length,
                **kwargs,
            )

        return function

    # If funct is None, it means the decorator is called with arguments (like @pyiron_job(...))
    if funct is None:
        return pyiron_job_function

    # If funct is not None, it means the decorator is called without parentheses (like @pyiron_job)
    else:
        # Assume this usage and handle the decorator like `pyiron_job_simple`
        def function(
            *args,
            pyiron_project: Project = Project("."),
            pyiron_resource_dict: dict = {},
            list_length: Optional[int] = None,
            **kwargs,
        ):
            resource_default_dict = {
                k: v.default for k, v in inspect.signature(Server).parameters.items()
            }
            return get_delayed_object(
                *args,
                python_function=funct,
                pyiron_project=pyiron_project,
                pyiron_resource_dict=pyiron_resource_dict,
                resource_default_dict=resource_default_dict,
                list_length=list_length,
                **kwargs,
            )

        return function



import importlib
import os
import pkgutil
import sys
import warnings

import pandas

import pyiron_base.storage.hdfio
from pyiron_base import state
from pyiron_base.database.performance import get_database_statistics
from pyiron_base.project.update.pyiron_base_03x_to_04x import pyiron_base_03x_to_04x

# we sometimes move classes between modules; this would break HDF storage,
# since objects save there the module path from which their classes can be
# imported.  We can work around this by defining here an explicit map that
# _to_object can use to find the new modules and update the HDF5 files
_MODULE_CONVERSION_DICT = {
    "pyiron_base.generic.datacontainer": "pyiron_base.storage.datacontainer",
    "pyiron_base.generic.inputlist": "pyiron_base.storage.inputlist",
    "pyiron_base.generic.flattenedstorage": "pyiron_base.storage.flattenedstorage",
    "pyiron_base.table.datamining": "pyiron_base.jobs.datamining",
}


def add_module_conversion(old: str, new: str):
    """
    Add a new module conversion.

    After setting up a conversion, call :meth:`.Project.maintenance.local.update_hdf_types` to rewrite the HDF5 files to
    make this change and allow loading of previously saved objects.

    Args:
        old (str): path to module that previously defined objects in storage
        new (str): path to module that should be imported instead
    Raises:
        ValueError: if an entry for `old` already exists and does not point to `new`.
    """
    if old not in _MODULE_CONVERSION_DICT:
        _MODULE_CONVERSION_DICT[old] = new
    elif _MODULE_CONVERSION_DICT[old] != new:
        raise ValueError(
            f"Module path '{old}' already found in conversion dict, pointing to '{_MODULE_CONVERSION_DICT[old]}'!"
        )


class Maintenance:
    """
    The purpose of maintenance class is to provide
    some measures of perfomance for pyiron, whether local to the project
    or global (describing the status of pyiron on the running machine)
    """

    def __init__(self, project):
        """
        Args:
            (project): pyiron project to do maintenance on
        """
        self._project = project
        self._global = GlobalMaintenance()
        self._update = UpdateMaintenance(self._project)
        self._local = LocalMaintenance(self._project)

    @property
    def global_status(self):
        return self._global

    @property
    def update(self):
        return self._update

    @property
    def local(self):
        return self._local

    @staticmethod
    def get_repository_status() -> pandas.DataFrame:
        """
        Finds the hashes and versions for every `pyiron` module available.

        Returns:
            pandas.DataFrame: The name of each module and the hash and version for its current git head.
        """
        from git import InvalidGitRepositoryError, Repo

        module_names = [
            name for _, name, _ in pkgutil.iter_modules() if name.startswith("pyiron")
        ]

        report = pandas.DataFrame(
            columns=["Module", "Git head", "Version"], index=range(len(module_names))
        )
        for i, name in enumerate(module_names):
            module = importlib.import_module(name)
            try:
                repo = Repo(os.path.dirname(os.path.dirname(module.__file__)))
            except InvalidGitRepositoryError:
                hash_ = "Not a repo"
            else:
                try:
                    hash_ = repo.head.reference.commit.hexsha
                except (ValueError, TypeError):
                    hash_ = "Error while resolving sha"
            if hasattr(module, "__version__"):
                version = module.__version__
            else:
                version = "not defined"
            report.loc[i] = [name, hash_, version]

        return report


class LocalMaintenance:
    def __init__(self, project):
        self._project = project

    def defragment_storage(
        self,
        recursive: bool = True,
        progress: bool = True,
        **kwargs: dict,
    ):
        """
        Rewrite the hdf5 files of jobs.  This can free up unused space.

        By default iterate recursively over the jobs within the current
        project.  This can be controlled with `recursive` and `kwargs`.

        Args:
            recursive (bool): search subprojects [True/False] - True by default
            progress (bool): if True (default), add an interactive progress bar to the iteration
            **kwargs (dict): Optional arguments for filtering with keys matching the project database column name
                            (eg. status="finished"). Asterisk can be used to denote a wildcard, for zero or more
                            instances of any character
        """
        for job in self._project.iter_jobs(
            recursive=recursive, progress=progress, convert_to_object=False, **kwargs
        ):
            hdf = job.project_hdf5
            hdf.rewrite_hdf5(job.name)

    def update_hdf_types(
        self,
        recursive: bool = True,
        progress: bool = True,
        **kwargs: dict,
    ):
        """
        Rewrite TYPE fields in hdf5 files for renamed modules.

        New module conversions can be added with
        :func:`.add_module_conversion(old, new)`.  This method will then
        consider all objects previously imported from `old` to be imported from
        `new`.

        Args:
            recursive (bool): search subprojects [True/False] - True by default
            progress (bool): if True (default), add an interactive progress bar to the iteration
            **kwargs (dict): Optional arguments for filtering with keys matching the project database column name
                            (eg. status="finished"). Asterisk can be used to denote a wildcard, for zero or more
                            instances of any character
        """

        def recurse(hdf):
            contents = hdf.list_all()
            for group in contents["groups"]:
                recurse(hdf[group])
            if "TYPE" in contents["nodes"]:
                (
                    module_path,
                    class_name,
                ) = pyiron_base.storage.hdfio._extract_module_class_name(hdf["TYPE"])
                if module_path in _MODULE_CONVERSION_DICT:
                    new_module_path = _MODULE_CONVERSION_DICT[module_path]
                    hdf["TYPE"] = f"<class '{new_module_path}.{class_name}'>"

        for job in self._project.iter_jobs(
            recursive=recursive, progress=progress, convert_to_object=False, **kwargs
        ):
            hdf = job.project_hdf5
            recurse(hdf)

        def fix_project_data(pr):
            try:
                hdf = pr.create_hdf(pr.path, "project_data")["../data"]
                recurse(hdf)
            except ValueError:
                # in case project data does not exist yet
                pass

        fix_project_data(self._project)
        for sub in self._project.iter_groups():
            fix_project_data(sub)

        self.update_pyiron_tables(recursive=recursive, progress=progress, **kwargs)

    def update_pyiron_tables(
        self,
        recursive: bool = True,
        progress: bool = True,
        **kwargs: dict,
    ):
        kwargs["hamilton"] = "PyironTable"
        for old, new in _MODULE_CONVERSION_DICT.items():
            sys.modules[old] = importlib.import_module(new)

        for job in self._project.iter_jobs(
            recursive=recursive, progress=progress, **kwargs
        ):
            job.to_hdf()


class UpdateMaintenance:
    def __init__(self, project):
        self._project = project

    def base_to_current(self, start_version: str, project=None):
        """Runs all updates for pyiron_base to reach the current version.

        Args:
            start_version(str): Version of pyiron_base in the mayor.minor[.patch] format from which to start applying
                the updates.
            project(None/project/list/str): The project(s) to be converted from 0.3 to 0.4 ; default: current project
                One may provide a pyiron Project, a list of pyiron Projects, or a string containing "all" or a valid
                path.
                If "all" is provided, pyiron tries to find all projects using the PROJECT_PATHS defined in the
                configuration.
        """
        mayor, minor = start_version.split(".")[0:2]
        if int(mayor) != 0:
            raise ValueError("Updates to version >0.x.y is not possible.")
        if int(minor) < 4:
            self.base_v0_3_to_v0_4(project)

    def base_v0_3_to_v0_4(self, project=None):
        """Update hdf files written with pyiron_base-0.3.x to pyiron_base-0.4.x

        pyiron_base<=0.3.9 has a bug that writes all arrays with dtype=object even
        numeric ones.  As a fix pyiron_base=0.4.0 introduces a conversion when reading
        such arrays, but does not automatically save them.  This conversion script
        simply goes over all jobs and rewrites their HDF5 files, since it's read with
        the correct dtype, this then writes this correct dtype.

        Args:
            project(None/project/list/str): The project(s) to be converted from 0.3 to 0.4 ; default: current project
                One may provide a pyiron Project, a list of pyiron Projects, or a string containing "all" or a valid
                path.
                If "all" is provided, pyiron tries to find all projects using the PROJECT_PATHS defined in the
                configuration.
        """
        if project is None:
            projects = [self._project]
        elif isinstance(project, list):
            projects = project
        elif project == "all":
            projects = [
                self._project.__class__(path)
                for path in state.settings.configuration["project_paths"]
            ]
        elif isinstance(project, str):
            if os.path.isdir(project):
                projects = [self._project.__class__(project)]
            else:
                raise ValueError(
                    f"{project} is a str but neither 'all' nor a directory."
                )
        else:
            projects = [project]

        if len(projects) == 0:
            warnings.warn(
                f"Provided project {project} lead to 0 projects to be converted."
            )

        for pr in projects:
            try:
                pyiron_base_03x_to_04x(pr)
            except ValueError as e:
                print(f"WARNING: Updating project {project} failed with {e}!")


class GlobalMaintenance:
    def __init__(self):
        """
        initialize the flag self._check_postgres, to control whether pyiron is
        set to communicate with a postgres database.
        """
        connection_string = state.database.sql_connection_string
        if "postgresql" not in connection_string:
            warnings.warn(
                """
                The database statistics is only available for a Postgresql database
                """
            )
            self._check_postgres = False
        else:
            self._check_postgres = True

    def get_database_statistics(self):
        if self._check_postgres:
            return get_database_statistics()
        else:
            raise RuntimeError(
                """
                The detabase statistics is only available for a Postgresql database
                """
            )
